import pathlib
import uuid

import cloudpathlib
import fsspec
import pyarrow.parquet
import pytest

import datasets
from molflux.datasets import load_dataset_from_store, save_dataset_to_store


@pytest.fixture(scope="module")
def fixture_mock_dataset() -> datasets.Dataset:
    data = {"col1": [1, 2], "col2": [3, 4]}
    return datasets.Dataset.from_dict(data)


@pytest.fixture(scope="module")
def fixture_mock_dataset_dict() -> datasets.DatasetDict:
    data = datasets.Dataset.from_dict({"col1": [1, 2], "col2": [3, 4]})
    return datasets.DatasetDict(
        {
            "train": data,
            "validation": data,
            "test": data,
        },
    )


@pytest.mark.parametrize(
    ("path", "format"),
    [
        # simple paths
        ("data.csv", "csv"),
        ("data.parquet", "parquet"),
        ("data.json", "json"),
        ("data", "disk"),
        # nested paths
        ("sub/dir/data.csv", "csv"),
        ("sub/dir/data.parquet", "parquet"),
        ("sub/dir/data.json", "json"),
        ("sub/dir/data", "disk"),
        # pathlib paths
        (pathlib.Path("data.csv"), "csv"),
        (pathlib.Path("sub/dir/data.csv"), "csv"),
        # compressed paths
        ("data.csv.gz", "csv"),
        # paths with potentially problematic characters
        ("2023-03-22T17:53:17.245386.parquet", "parquet"),
    ],
)
def test_save_dataset_locally(tmp_path, fixture_mock_dataset, path, format):
    """That can persist a Dataset locally in one of the supported file
    formats.

    We also test that missing parent directories are automatically created if
    needed.
    """
    dataset = fixture_mock_dataset
    path = tmp_path / path

    save_dataset_to_store(dataset, path=str(path), format=format)

    if format == "disk":
        assert path.is_dir()
    else:
        assert path.is_file()


@pytest.mark.parametrize(
    ("path", "format"),
    [
        # simple paths
        ("s3://data.csv", "csv"),
        ("s3://data.parquet", "parquet"),
        ("s3://data.json", "json"),
        ("s3://data", "disk"),
        # nested paths
        ("s3://sub/dir/data.csv", "csv"),
        ("s3://sub/dir/data.parquet", "parquet"),
        ("s3://sub/dir/data.json", "json"),
        ("s3://sub/dir/data", "disk"),
        # cloudpathlib paths
        (cloudpathlib.S3Path("s3://data.csv"), "csv"),
        (cloudpathlib.S3Path("s3://sub/dir/data.csv"), "csv"),
        # compressed paths
        ("s3://data.csv.gz", "csv"),
        # paths with potentially problematic characters
        ("s3://data/2023-03-22T17:53:17.245386.parquet", "parquet"),
    ],
)
def test_save_dataset_to_s3(
    fixture_mock_s3_client,
    fixture_mock_s3_filesystem,
    fixture_mock_dataset,
    path,
    format,
):
    """That can persist a Dataset to the cloud in one of the supported file
    formats.

    We also test that missing parent directories are automatically created if
    needed (even though technically speaking in cloud paths there is not really
    a concept of subdirectories).
    """
    dataset = fixture_mock_dataset

    # Prepare s3 bucket
    client = fixture_mock_s3_client
    root_bucket = str(uuid.uuid4())
    client.create_bucket(Bucket=root_bucket)

    # point at mock bucket
    cloud_path = str(path).replace("s3://", f"s3://{root_bucket}/")
    if isinstance(path, cloudpathlib.CloudPath):
        cloud_path = cloudpathlib.S3Path(cloud_path)  # type: ignore[assignment]

    # Save to s3
    fs = fixture_mock_s3_filesystem
    save_dataset_to_store(dataset, path=cloud_path, format=format, fs=fs)

    if not pathlib.Path(str(cloud_path)).suffix:
        assert fs.isdir(str(cloud_path))
    else:
        assert fs.isfile(str(cloud_path))


@pytest.mark.parametrize(
    ("path", "format", "filesystem"),
    [
        ("s3://my/bucket/data.parquet", "parquet", "file"),
        ("my/data.parquet", "parquet", "s3"),
    ],
)
def test_save_dataset_with_mismatched_filesystem_raises(
    fixture_mock_s3_client,
    fixture_mock_s3_filesystem,
    fixture_mock_dataset,
    path,
    format,
    filesystem,
):
    """That attempting to save a dataset to the cloud but providing a
    local filesystem (local) raises (and vice-versa)."""
    dataset = fixture_mock_dataset
    with pytest.raises(ValueError, match="Incompatible filesystem"):
        save_dataset_to_store(
            dataset,
            path=str(path),
            format=format,
            fs=fsspec.filesystem(filesystem),
        )


@pytest.mark.parametrize(
    ("filename", "wrong_format"),
    [
        # simple paths
        ("data.csv", "disk"),
        ("data.parquet", "json"),
        ("data.json", "parquet"),
        ("data", "csv"),
    ],
)
def test_save_dataset_with_mismatched_file_format_raises(
    tmp_path,
    fixture_mock_dataset,
    filename,
    wrong_format,
):
    """That an error is raised if a file format is provided that does not
    match the data filename.
    """
    dataset = fixture_mock_dataset
    path = tmp_path / filename

    with pytest.raises(ValueError, match="Mismatched file format"):
        save_dataset_to_store(dataset, path=str(path), format=wrong_format)


def test_save_dataset_to_not_implemented_format_raises(fixture_mock_dataset):
    """That attempting to save a dataset to an unimplemented format raises."""
    dataset = fixture_mock_dataset

    with pytest.raises(ValueError, match="Unsupported dataset format"):
        save_dataset_to_store(dataset, path="test", format="invalid")  # type: ignore[arg-type]


@pytest.mark.parametrize(
    "filename",
    ["data.csv", "data.parquet", "data.json", "data", "data.csv.gz"],
)
def test_dataset_file_format_is_not_required_and_is_autoresolved(
    tmp_path,
    fixture_mock_dataset,
    filename,
):
    """That can persist a Dataset locally in one of the supported file
    formats without explicitly specifying the file format."""

    dataset = fixture_mock_dataset
    path = tmp_path / filename

    save_dataset_to_store(dataset, path=str(path))

    if not path.suffixes:
        assert path.is_dir()
    else:
        assert path.is_file()


@pytest.mark.parametrize(
    "compression",
    ["NONE", "SNAPPY", "GZIP", "BROTLI", "LZ4", "ZSTD"],
)
def test_save_dataset_as_compressed_parquet(
    tmp_path,
    fixture_mock_dataset,
    compression,
):
    """That can save datasets as compressed parquet files.

    Kwargs are forwarded to the backend huggingface .to_parquet() method.

    References:
        https://huggingface.co/docs/datasets/package_reference/main_classes#datasets.Dataset.to_parquet
        https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetWriter.html
    """
    dataset = fixture_mock_dataset

    path = tmp_path / "compressed.parquet"
    save_dataset_to_store(dataset, path=str(path), compression=compression)
    compression_metadata = (
        pyarrow.parquet.ParquetFile(path).metadata.row_group(0).column(0).compression
    )

    if compression == "NONE":
        assert compression_metadata == "UNCOMPRESSED"
    else:
        assert compression_metadata == compression


@pytest.mark.parametrize(
    ("path", "format"),
    [
        # simple paths
        ("data", "csv"),
        ("data", "parquet"),
        ("data", "json"),
        ("data", "disk"),
        # nested paths
        ("sub/dir/data", "csv"),
        ("sub/dir/data", "parquet"),
        ("sub/dir/data", "json"),
        ("sub/dir/data", "disk"),
    ],
)
def test_save_dataset_dict_locally(tmp_path, fixture_mock_dataset_dict, path, format):
    """That can persist a DatasetDict locally in one of the supported file
    formats.

    We also test that missing parent directories are automatically created if
    needed.
    """
    dataset_dict = fixture_mock_dataset_dict
    path = tmp_path / path

    save_dataset_to_store(dataset_dict, path=str(path), format=format)
    assert path.is_dir()

    # The split name formatting is technically an implementation detail
    # do not hesitate to adapt if code changes
    for split_name in dataset_dict:
        if format == "disk":
            expected_split_path = path / split_name
            assert expected_split_path.is_dir()
        else:
            expected_split_path = path / f"{split_name}.{format}"
            assert expected_split_path.is_file()


def test_save_dataset_dict_defaults_to_parquet(tmp_path, fixture_mock_dataset_dict):
    """That the default persistence format a DatasetDict is saved as is parquet.

    This is because we believe that .parquet is the most sensible persistence
    format for general usage
    """
    dataset_dict = fixture_mock_dataset_dict
    path = tmp_path / "data"

    save_dataset_to_store(dataset_dict, path=str(path))
    assert path.is_dir()

    # The split name formatting is technically an implementation detail
    # do not hesitate to adapt if code changes
    for split_name in dataset_dict:
        expected_split_path = path / f"{split_name}.parquet"
        assert expected_split_path.is_file()


def test_non_directory_like_path_for_dataset_dict_raises(
    tmp_path,
    fixture_mock_dataset_dict,
):
    """That DatasetDicts can only be saved to directories."""
    dataset_dict = fixture_mock_dataset_dict
    path = tmp_path / "data.parquet"
    with pytest.raises(ValueError, match="Invalid path"):
        save_dataset_to_store(dataset_dict, path=str(path), format="parquet")


@pytest.mark.parametrize(
    ("path", "format"),
    [
        ("sub/dir/data.csv", "csv"),
        ("sub/dir/data.parquet", "parquet"),
        ("sub/dir/data.json", "json"),
        ("sub/dir/data", "disk"),
    ],
)
def test_round_trip_dataset_io(tmp_path, fixture_mock_dataset, path, format):
    """That can save and load Datasets in a consistent fashion."""
    dataset = fixture_mock_dataset
    path = tmp_path / path

    save_dataset_to_store(dataset, path=str(path), format=format)
    loaded_dataset = load_dataset_from_store(str(path), format=format)

    assert loaded_dataset.column_names == dataset.column_names
    assert loaded_dataset.shape == dataset.shape


@pytest.mark.parametrize(
    ("path", "format"),
    [
        ("sub/dir/data", "csv"),
        ("sub/dir/data", "parquet"),
        ("sub/dir/data", "json"),
        ("sub/dir/data", "disk"),
        # test that default format is interpreted consistently
        ("sub/dir/data", None),
    ],
)
def test_round_trip_dataset_dict_io(tmp_path, fixture_mock_dataset_dict, path, format):
    """That can save and load DatasetDicts in a consistent fashion."""
    dataset_dict = fixture_mock_dataset_dict
    path = tmp_path / path

    save_dataset_to_store(dataset_dict, path=str(path), format=format)
    loaded_dataset_dict = load_dataset_from_store(str(path), format=format)

    assert loaded_dataset_dict.column_names == dataset_dict.column_names
    assert loaded_dataset_dict.shape == dataset_dict.shape
    assert list(loaded_dataset_dict.keys()) == list(dataset_dict.keys())



import pytest

import molflux.splits.catalogue
from molflux.splits.catalogue import (
    get_splitting_strategy_cls,
    list_splitting_strategies,
    register_splitting_strategy,
)
from molflux.splits.strategy import SplittingStrategy

catalogue_obj = molflux.splits.catalogue.SPLITTING_STRATEGIES_CATALOGUE


def test_catalogue_is_filled():
    """That the splitting strategies catalogue has been filled."""
    catalogue = list_splitting_strategies()
    assert len(catalogue)


def test_list_splitting_strategies_returns_sorted_names():
    """That strategy names in the catalogue are returned sorted."""
    catalogue = list_splitting_strategies()
    assert sorted(catalogue.keys()) == list(catalogue.keys())
    for strategies in catalogue.values():
        assert sorted(strategies) == strategies


def test_list_splitting_strategies_lists_names_of_strategies_in_catalogue():
    """That all strategies existing in the catalogue are returned by list_splitting_strategies."""
    view = list_splitting_strategies()
    names_in_view = [name for names in view.values() for name in names]
    assert sorted(names_in_view) == sorted(catalogue_obj.keys())


def test_register_strategy(monkeypatch):
    """That can register a new splitting strategy in the catalogue."""
    # set up a mock empty catalogue to leave real one untouched
    monkeypatch.setattr(molflux.splits.catalogue, "SPLITTING_STRATEGIES_CATALOGUE", {})

    assert "pytest_strategy" not in list_splitting_strategies()

    @register_splitting_strategy(kind="testing", name="pytest_strategy")
    class PytestStrategy: ...

    assert "testing" in list_splitting_strategies()
    assert "pytest_strategy" in list_splitting_strategies()["testing"]


def test_get_strategy_not_in_catalogue_raises_not_implemented_error():
    """That getting a non-existent strategy raises a NotImplementedError."""
    strategy_name = "non-existent-strategy"
    with pytest.raises(
        NotImplementedError,
        match=f"Splitting strategy {strategy_name!r} is not available",
    ):
        get_splitting_strategy_cls(strategy_name)


def test_get_strategy_cls_returns_splitting_strategy():
    """That getting a cls from the catalogue returns a class implementing the protocol."""
    strategy_name = "linear_split"
    splitting_strategy_cls = get_splitting_strategy_cls(strategy_name)
    assert isinstance(splitting_strategy_cls, SplittingStrategy)



from molflux.datasets.load import load_from_dicts

representative_dataset_name = "esol"


def test_returns_dataset_dict():
    """That loading from a dict returns a Dataset."""
    name = representative_dataset_name
    config = {
        "name": name,
        "config": {},
    }
    dataset_map = load_from_dicts([config])
    assert isinstance(dataset_map, dict)



import warnings
from collections.abc import Iterable, Mapping
from typing import Any

import datasets
import molflux.modelzoo
from datasets import Dataset, DatasetDict
from molflux.core.environment import save_python_environment
from molflux.core.featurisation.metadata import save_featurisation_metadata
from molflux.core.typing import AllowedInferenceMethods, PathLike
from molflux.modelzoo import Model


def save_model(
    model: Model,
    path: PathLike,
    featurisation_metadata: dict[str, Any] | None,
) -> str:
    """Packages a trained model to disk into a standard format.

    This standardised format ensures that models can be deployed in production
    and used seamlessly in a variety of downstream tools.

    .. code-block: shell
        model/
        ├── model_config.json
        ├── model_artefacts/
        |   └── model.pkl
        ├── featurisation_metadata.json
        └── requirements.txt

    Args:
        model: The model to save.
        path: The path under which to save the model.
        featurisation_metadata: The featurisation metadata for the features the
            model was trained on.

    Returns:
        A reference to the path the model was saved to.
    """

    # save model to store
    molflux.modelzoo.save_to_store(key=path, model=model)

    # save featurisation config alongside of it
    if not featurisation_metadata:
        warnings.warn(
            "You have saved a model but did not track its featurisation metadata. "
            "Users might not be able to automatically inference your model from unfeaturised inputs",
            UserWarning,
            stacklevel=1,
        )

    save_featurisation_metadata(
        featurisation_metadata=featurisation_metadata,
        path=path,
    )

    # save python environment metadata alongside of it
    save_python_environment(path=path)

    return str(path)


def load_model(path: str) -> Model:
    """Loads a model saved to disk."""
    return molflux.modelzoo.load_from_store(key=path)


def get_references(model: Model, fold: DatasetDict) -> DatasetDict:
    """Extracts the model ground truth references from a fold."""
    return fold.select_columns(column_names=model.y_features)


def get_inputs(model: Model, fold: DatasetDict) -> DatasetDict:
    """Extracts the model's input features from a fold."""
    return fold.select_columns(column_names=model.x_features)


def _dict_of_iterables_to_iterable_of_dicts(
    d: Mapping[str, Iterable[Any]],
) -> Iterable[dict[str, Any]]:
    """Split dictionary of iterables into iterable of dictionaries.

    References:
        https://stackoverflow.com/a/1780295
    """
    return map(
        dict,
        zip(*[[(k, v) for v in value] for k, value in d.items()], strict=False),
    )


def predict(
    model: Model,
    fold: DatasetDict,
    prediction_method: AllowedInferenceMethods = "predict",
    prediction_kwargs: dict[str, dict[str, Any]] | None = None,
) -> DatasetDict | tuple[DatasetDict, ...]:
    """Calculates model predictions for each split in a fold.

    Model predictions follow the molflux.modelzoo predictions output format, where
    predictions are returned for each predicted task.

    Args:
        model: The model to use for predictions.
        fold: The fold of prediction inputs.
        prediction_method: The method to use for predictions. Should be one of
            ["predict", "predict_proba", "predict_with_prediction_interval",
            "predict_with_std", "sample"]. Defaults to "predict".
        prediction_kwargs: A dictionary of extra prediction kwargs to use for
            each split in the fold.

    Returns:
        DatasetDict(s) of prediction results.
    """

    # Initialise the dictionary of prediction outputs for each input split dataset
    # ensuring same ordering of split names for more intuitive behaviour
    fold_prediction_outputs: dict[str, tuple[DatasetDict, ...]] = dict.fromkeys(
        fold,
        (DatasetDict(),),
    )

    for split_name, dataset in fold.items():
        # make sure malformed kwargs do not go through silently
        if prediction_kwargs and split_name not in prediction_kwargs:
            raise KeyError(f"Please provide prediction kwargs for split {split_name!r}")

        kwargs = prediction_kwargs[split_name] if prediction_kwargs else {}
        prediction_results = inference(
            model,
            dataset=dataset,
            prediction_method=prediction_method,
            prediction_kwargs=kwargs,
        )

        if not isinstance(prediction_results, tuple):
            prediction_results = (prediction_results,)

        fold_prediction_outputs[split_name] = prediction_results

    # An iterable of dicts of Datasets, one for each prediction_method item
    # {"train": (y, ...), "validation": (y, ...), "test": (y, ...)}
    # ->
    # (DatasetDict({"train": y, "validation": y, "test": y}), ...)
    out = tuple(
        DatasetDict(fold_predictions)
        for fold_predictions in _dict_of_iterables_to_iterable_of_dicts(
            fold_prediction_outputs,
        )
    )
    return out[0] if len(out) == 1 else out


def inference(
    model: Model,
    dataset: Dataset,
    prediction_method: AllowedInferenceMethods = "predict",
    prediction_kwargs: dict[str, Any] | None = None,
) -> Dataset | tuple[Dataset, ...]:
    """Calculates model predictions on a single dataset."""

    predictor = getattr(model, prediction_method)
    prediction_results = predictor(dataset, **(prediction_kwargs or {}))

    if isinstance(prediction_results, tuple):
        return tuple(datasets.Dataset.from_dict(d) for d in prediction_results)
    else:
        return datasets.Dataset.from_dict(prediction_results)



"""Calibration gap."""

import logging
from typing import Any

import evaluate
import numpy as np
from scipy.stats import norm

import datasets
from molflux.metrics.bases import UncertaintyMetric
from molflux.metrics.typing import ArrayLike, MetricResult
from molflux.metrics.uncertainty.utils import _estimate_standard_deviation

logger = logging.getLogger(__name__)


_DESCRIPTION = """
The calibration gap is the unsigned area between a models calibration curve and a perfectly calibrated model (y=x).
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: array of mean predictions.
    references: Ground truth (correct) target values.
    standard_deviations: Standard deviation predicted from a model with uncertainty.
        Either standard deviations or prediction_intervals should be provided.
    prediction_intervals: Prediction intervals of lower and upper bounds.
        Either standard deviations or prediction_intervals should be provided.
    confidence (optional): float indicating what proportion of data lies between
        lower and upper bounds. Defaults to 0.9.
    num_thresholds (optional): int giving the number of thresholds at which to
        assess the calibration error. Defaults to 1000.

Returns:
    calibration_gap: unsigned area between the calibration curve and y=x.

Examples:
    >>> from molflux.metrics import load_metric
    >>> m = load_metric("calibration_gap")
    >>> ref = [0, 1, 2, 3, 4]
    >>> pred = [0.5, 0.8, 1.2, 2.4, 4.5]
    >>> lower_bound = [0.1, 0.1, 0.2, 2, 3]
    >>> upper_bound = [1, 1.5, 1.8, 2.8, 5.7]
    >>> prediction_intervals = list(zip(lower_bound, upper_bound))
    >>> m.compute(predictions=pred, references=ref, prediction_intervals=prediction_intervals)
    {'calibration_gap': 0.1314763799612285}
"""

_CITATION = """
@inproceedings{kuleshov2018accurate,
  title={Accurate uncertainties for deep learning using calibrated regression},
  author={Kuleshov, Volodymyr and Fenner, Nathan and Ermon, Stefano},
  booktitle={International conference on machine learning},
  pages={2796--2804},
  year={2018},
  organization={PMLR}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class CalibrationGap(UncertaintyMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("float64"),
                    "references": datasets.Value("float64"),
                },
            ),
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        standard_deviations: ArrayLike | None = None,
        prediction_intervals: ArrayLike | None = None,
        confidence: float = 0.9,
        num_thresholds: int = 100,
        **kwargs: Any,
    ) -> MetricResult:
        if standard_deviations is None:
            if prediction_intervals is None:
                raise ValueError(
                    "Please provide either standard deviation, or prediction intervals",
                )
            lower_bound, upper_bound = zip(*prediction_intervals, strict=False)
            standard_deviations = _estimate_standard_deviation(
                lower_bound,
                upper_bound,
                confidence,
            )

        assert len(predictions) == len(references)
        pred_cdf = norm.cdf(references, loc=predictions, scale=standard_deviations)
        # expected quantile
        rho_s = np.linspace(0, 1, num_thresholds)
        # empirical frequency of predicted values falling in the rho_s quantile
        p_s = np.array([np.mean(pred_cdf < rho) for rho in rho_s])
        area = np.trapz(np.abs(rho_s - p_s), x=rho_s)
        return {self.tag: area}



import difflib
import functools
import inspect
import logging
from collections import defaultdict
from collections.abc import Callable
from importlib.metadata import EntryPoint, entry_points

from molflux.features.naming import camelcase_to_snakecase
from molflux.features.representation import Representation

logger = logging.getLogger(__name__)

RepresentationT = type[Representation]

NAMESPACE = "molflux.features.plugins."

# This is where entrypoints will be registered {<name>: <Entrypoint>}
REPRESENTATIONS_CATALOGUE: dict[str, EntryPoint] = {}


@functools.lru_cache
def fill_catalogue() -> None:
    """Fills the catalogue with entrypoints.

    Only ever call this once for performance reasons.
    """
    entrypoints = [
        entrypoint
        for namespace, entrypoints in entry_points().items()
        if namespace.startswith(NAMESPACE)
        for entrypoint in entrypoints
    ]
    for entrypoint in entrypoints:
        put_representation_entrypoint(representation_entrypoint=entrypoint)


def put_representation_entrypoint(representation_entrypoint: EntryPoint) -> None:
    """Puts a representation entrypoint in the catalogue."""
    representation_name = camelcase_to_snakecase(representation_entrypoint.name)

    if representation_name in REPRESENTATIONS_CATALOGUE:
        if representation_entrypoint == REPRESENTATIONS_CATALOGUE[representation_name]:
            pass
        else:
            raise KeyError(f"Duplicate representation: {representation_name!r}")

    REPRESENTATIONS_CATALOGUE[representation_name] = representation_entrypoint


def get_representation_entrypoint(representation_name: str) -> EntryPoint:
    """Returns a specific representation entrypoint from the catalogue.

    The entrypoint is keyed only by name. This allows to decouple from
    eventual reshuffles of entrypoints across groups.
    """

    representation_entrypoint = REPRESENTATIONS_CATALOGUE.get(representation_name)
    if representation_entrypoint is None:
        msg = f"Representation {representation_name!r} is not available."
        similar = difflib.get_close_matches(
            representation_name,
            REPRESENTATIONS_CATALOGUE.keys(),
        )
        if similar:
            msg += f" You might be looking for one of these: {similar}"
        raise NotImplementedError(msg)
    return representation_entrypoint


def get_representation_cls(representation_name: str) -> RepresentationT:
    entrypoint = get_representation_entrypoint(representation_name)

    # Might raise (e.g. missing dependency extras)
    representation_cls = entrypoint.load()

    if not isinstance(representation_cls, Representation):
        raise NotImplementedError(
            f"Plugin {entrypoint.value} does not implement Representation protocol.",
        )

    return representation_cls  # type: ignore[return-value]


def list_representations() -> dict[str, list[str]]:
    """List all available representations.

    The catalogue is returned as a view of representation names keyed by
    kind.
    """

    view: dict[str, list[str]] = defaultdict(list)

    for (
        representation_name,
        representation_entrypoint,
    ) in REPRESENTATIONS_CATALOGUE.items():
        representation_kind = representation_entrypoint.group.split(".")[-1]
        view[representation_kind].append(representation_name)

    return dict(sorted(view.items()))


def register_representation(kind: str, name: str) -> Callable:
    """Registers a custom representation in the entrypoints catalogue.

    Examples:
        .. code-block:: python

           @register_representation(kind="custom", name="my_representation")
           class CustomRepresentation:
               ...
    """

    def wrapper(representation_cls: RepresentationT) -> RepresentationT:
        module = inspect.getmodule(representation_cls).__name__  # type: ignore
        value = f"{module}:{representation_cls.__name__}"
        group = NAMESPACE + kind
        entrypoint = EntryPoint(name=name, value=value, group=group)
        put_representation_entrypoint(representation_entrypoint=entrypoint)
        return representation_cls

    return wrapper



import os
from abc import ABC, abstractmethod
from dataclasses import asdict
from typing import Any, Dict, List, Optional, Protocol

import joblib
import numpy as np

import datasets
from molflux.modelzoo.errors import NotTrainedError
from molflux.modelzoo.model import ClassificationMixin, ModelBase, _ModelConfigT
from molflux.modelzoo.typing import Classes, PredictionResult
from molflux.modelzoo.utils import (
    get_concatenated_array,
    pick_features,
    validate_features,
)


class _SKLearnBaseAPI(Protocol):
    """The interface of a typical sklearn model.

    We are formalising it ourselves here because sklearn doesn't have a base
    class that defines these two methods.
    """

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray | None = None,
        **kwargs: Any,
    ) -> Any: ...

    def predict(self, X: np.ndarray, **kwargs: Any) -> np.ndarray: ...


class SKLearnModelBase(ModelBase[_ModelConfigT], ABC):
    """Base model for all models based on the sklearn API"""

    @abstractmethod
    def _instantiate_model(self) -> _SKLearnBaseAPI: ...

    def _train(
        self,
        train_data: datasets.Dataset,
        **kwargs: Any,
    ) -> Any:
        # validate y features as well
        validate_features(train_data, self.y_features)

        x_data = pick_features(train_data, self.x_features)
        X = get_concatenated_array(x_data, self.x_features)

        y_data = pick_features(train_data, self.y_features)
        y = get_concatenated_array(y_data, self.y_features)

        # reshape (n,1) column-vector into (n,) 1d array, as expected by sklearn
        is_column_vector = (len(y.shape) > 1) and (y.shape[1] == 1)
        if is_column_vector:
            y = y.ravel()

        # instantiate model
        self.model = self._instantiate_model()

        # train
        self.model.fit(X, y, **kwargs)

    def _predict(
        self,
        data: datasets.Dataset,
        **kwargs: Any,
    ) -> PredictionResult:
        display_names = self._predict_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        X = get_concatenated_array(data, self.x_features)
        y_predict: np.ndarray = self.model.predict(X, **kwargs)
        # this could be of several shapes / types:
        # - singletask: (n_samples,)
        # - multitask: (n_tasks, n_samples)
        # lets align...
        if np.ndim(y_predict) == 1:
            y_predict = np.expand_dims(y_predict, axis=1)

        n_samples, n_tasks = np.shape(y_predict)

        # Check that output tasks match what specified
        if n_tasks != len(self.y_features):
            raise ValueError(
                f"Predictions do not have expected number of tasks: {np.shape(y_predict)!r}",
            )

        # (n_tasks, n_samples)
        return {
            display_name: task_predictions.tolist()
            for display_name, task_predictions in zip(
                display_names,
                y_predict.T,
                strict=False,
            )
        }

    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    def as_dir(self, directory: str) -> None:
        """Serialises a pre-trained model in a directory."""

        if self.model is None:
            raise NotTrainedError

        filename = os.path.join(directory, "model.joblib")
        joblib.dump(self.model, filename=filename, compress=0)

    def from_dir(self, directory: str) -> None:
        """Deserialises the backend model object stored in a given directory.

        This should undo the process defined in `self.as_dir()`.
        """

        try:
            filename = os.path.join(directory, "model.joblib")
            self.model = joblib.load(filename)
            return
        except FileNotFoundError as err:
            raise FileNotFoundError(
                f"Could not find scikit-learn model binaries: expected {filename!r}",
            ) from err


class _SKLearnClassifierAPI(Protocol):
    """The interface of a typical sklearn classifier.

    We are formalising it ourselves here because sklearn doesn't have a formal
    class that defines it
    """

    classes_: list[str]

    def predict_proba(self, X: np.ndarray, **kwargs: Any) -> np.ndarray: ...


class SKLearnClassificationMixin(ClassificationMixin):
    model: _SKLearnClassifierAPI
    tag: str

    @property
    def classes(self) -> Classes:
        if not hasattr(self.model, "classes_"):
            return {}

        sklearn_classes = self.model.classes_
        # this could be of several shapes / types
        # - singletask: (n_classes,)
        # - multitask: (n_tasks, n_classes)

        is_multitask = np.ndim(sklearn_classes) == 2
        classes = sklearn_classes if is_multitask else [sklearn_classes]  # type: ignore[list-item]

        return {
            task: classes.tolist()  # type: ignore[attr-defined]
            for task, classes in zip(self.y_features, classes, strict=False)
        }

    def _predict_proba(self, data: datasets.Dataset, **kwargs: Any) -> PredictionResult:
        display_names = self._predict_proba_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        X = get_concatenated_array(data, self.x_features)
        y_predict: np.ndarray = self.model.predict_proba(X, **kwargs)
        # this could be of several shapes / types:
        # - singletask: (n_samples, n_classes)
        # - multitask: (n_tasks, n_samples, n_classes)
        # lets align...
        if np.ndim(y_predict) == 2:
            y_predict = np.expand_dims(y_predict, axis=0)

        n_tasks, n_samples, n_classes = np.shape(y_predict)

        # Check that output tasks match what specified
        if n_tasks != len(self.y_features):
            raise ValueError(
                f"Predictions do not have expected number of tasks: {np.shape(y_predict)!r}",
            )

        results = {}
        for task, task_predictions, display_name in zip(
            self.y_features,
            y_predict,
            display_names,
            strict=False,
        ):
            n_samples, n_classes = np.shape(task_predictions)

            # Check that output classes match what specified
            if n_classes != len(self.classes[task]):
                raise ValueError(
                    f"Predictions do not have expected number of classes: {np.shape(task_predictions)!r}",
                )

            results[display_name] = task_predictions.tolist()

        # (n_tasks, n_samples, n_classes)
        return results



from collections.abc import Callable

import numpy as np
from numpy.random import RandomState
from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import SKLearnModelBase

try:
    from xgboost.sklearn import TrainingCallback, _SklObjective
    from xgboost.sklearn import XGBRegressor as XGBR
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("xgboost", e) from None


_DESCRIPTION = """
Implementation of the scikit-learn API for XGBoost regression.

Extreme Gradient Boosting, or XGBoost for short is an efficient
open-source implementation of the gradient boosting algorithm.

Gradient boosting refers to a class of ensemble machine learning
algorithms that can be used for classification or regression predictive
modeling problems.

Ensembles are constructed from decision tree models. Trees are added one
at a time to the ensemble and fit to correct the prediction errors made
by prior models. This is a type of ensemble machine learning model referred
to as boosting.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
n_estimators: int = 100
    Number of estimators, equivalent to the number of boosting rounds.
max_depth :  Optional[int]
    Maximum tree depth for base learners.
max_leaves : Optional[int]
    Maximum number of leaves; 0 indicates no limit.
max_bin : Optional[int]
    If using histogram-based algorithm, maximum number of bins per feature
grow_policy :
    Tree growing policy. 0: favor splitting at nodes closest to the node, i.e. grow
    depth-wise. 1: favor splitting at nodes with highest loss change.
learning_rate : Optional[float]
    Boosting learning rate (xgb's "eta")
verbosity : Optional[int]
    The degree of verbosity. Valid values are 0 (silent) - 3 (debug).
objective : {_SklObjective}
    Specify the learning task and the corresponding learning objective or
    a custom objective function to be used (see note below).
booster: Optional[str]
    Specify which booster to use: gbtree, gblinear or dart.
tree_method: Optional[str]
    Specify which tree method to use.  Default to auto.  If this parameter is set to
    default, XGBoost will choose the most conservative option available.  It's
    recommended to study this option from the parameters document :doc:`tree method
    </treemethod>`
n_jobs : Optional[int]
    Number of parallel threads used to run xgboost.  When used with other Scikit-Learn
    algorithms like grid search, you may choose which algorithm to parallelize and
    balance the threads.  Creating thread contention will significantly slow down both
    algorithms.
gamma : Optional[float]
    (min_split_loss) Minimum loss reduction required to make a further partition on a
    leaf node of the tree.
min_child_weight : Optional[float]
    Minimum sum of instance weight(hessian) needed in a child.
max_delta_step : Optional[float]
    Maximum delta step we allow each tree's weight estimation to be.
subsample : Optional[float]
    Subsample ratio of the training instance.
sampling_method :
    Sampling method. Used only by `gpu_hist` tree method.
      - `uniform`: select random training instances uniformly.
      - `gradient_based` select random training instances with higher probability when
        the gradient and hessian are larger. (cf. CatBoost)
colsample_bytree : Optional[float]
    Subsample ratio of columns when constructing each tree.
colsample_bylevel : Optional[float]
    Subsample ratio of columns for each level.
colsample_bynode : Optional[float]
    Subsample ratio of columns for each split.
reg_alpha : Optional[float]
    L1 regularization term on weights (xgb's alpha).
reg_lambda : Optional[float]
    L2 regularization term on weights (xgb's lambda).
scale_pos_weight : Optional[float]
    Balancing of positive and negative weights.
base_score : Optional[float]
    The initial prediction score of all instances, global bias.
random_state : Optional[Union[numpy.random.RandomState, int]]
    Random number seed.
    .. note::
       Using gblinear booster with shotgun updater is nondeterministic as
       it uses Hogwild algorithm.
missing : Optional[float], default None
    Value in the data which needs to be present as a missing value.
num_parallel_tree: Optional[int]
    Used for boosting random forest.
monotone_constraints : Optional[Union[Dict[str, int], str]]
    Constraint of variable monotonicity.  See :doc:`tutorial </tutorials/monotonic>`
    for more information.
interaction_constraints : Optional[Union[str, List[Tuple[str]]]]
    Constraints for interaction representing permitted interactions.  The
    constraints must be specified in the form of a nested list, e.g. ``[[0, 1], [2,
    3, 4]]``, where each inner list is a group of indices of features that are
    allowed to interact with each other.  See :doc:`tutorial
    </tutorials/feature_interaction_constraint>` for more information
importance_type: Optional[str]
    The feature importance type for the feature_importances\\_ property:
    * For tree model, it's either "gain", "weight", "cover", "total_gain" or
      "total_cover".
    * For linear model, only "weight" is defined and it's the normalized coefficients
      without bias.
gpu_id : Optional[int]
    Device ordinal.
validate_parameters : Optional[bool]
    Give warnings for unknown parameter.
predictor : Optional[str]
    Force XGBoost to use specific predictor, available choices are [cpu_predictor,
    gpu_predictor].
enable_categorical : bool
    .. note:: This parameter is experimental
    Experimental support for categorical data.  When enabled, cudf/pandas.DataFrame
    should be used to specify categorical data type.  Also, JSON/UBJSON
    serialization format is required.
feature_types : FeatureTypes
    Used for specifying feature types without constructing a dataframe. See
    :py:class:`DMatrix` for details.
max_cat_to_onehot : Optional[int]
    .. note:: This parameter is experimental
    A threshold for deciding whether XGBoost should use one-hot encoding based split
    for categorical data.  When number of categories is lesser than the threshold
    then one-hot encoding is chosen, otherwise the categories will be partitioned
    into children nodes.  Only relevant for regression and binary classification.
    See :doc:`Categorical Data </tutorials/categorical>` for details.
eval_metric : Optional[Union[str, List[str], Callable]]
    Metric used for monitoring the training result and early stopping.  It can be a
    string or list of strings as names of predefined metric in XGBoost (See
    doc/parameter.rst), one of the metrics in :py:mod:`sklearn.metrics`, or any other
    user defined metric that looks like `sklearn.metrics`.
    If custom objective is also provided, then custom metric should implement the
    corresponding reverse link function.
    Unlike the `scoring` parameter commonly used in scikit-learn, when a callable
    object is provided, it's assumed to be a cost function and by default XGBoost will
    minimize the result during early stopping.
    For advanced usage on Early stopping like directly choosing to maximize instead of
    minimize, see :py:obj:`xgboost.callback.EarlyStopping`.
    See :doc:`Custom Objective and Evaluation Metric </tutorials/custom_metric_obj>`
    for more.
    .. note::
         This parameter replaces `eval_metric` in :py:meth:`fit` method.  The old one
         receives un-transformed prediction regardless of whether custom objective is
         being used.
    .. code-block:: python
        from sklearn.datasets import load_diabetes
        from sklearn.metrics import mean_absolute_error
        X, y = load_diabetes(return_X_y=True)
        reg = xgb.XGBRegressor(
            tree_method="hist",
            eval_metric=mean_absolute_error,
        )
        reg.fit(X, y, eval_set=[(X, y)])
early_stopping_rounds : Optional[int]
    Activates early stopping. Validation metric needs to improve at least once in
    every **early_stopping_rounds** round(s) to continue training.  Requires at least
    one item in **eval_set** in :py:meth:`fit`.
    The method returns the model from the last iteration (not the best one).  If
    there's more than one item in **eval_set**, the last entry will be used for early
    stopping.  If there's more than one metric in **eval_metric**, the last metric
    will be used for early stopping.
    If early stopping occurs, the model will have three additional fields:
    :py:attr:`best_score`, :py:attr:`best_iteration` and
    :py:attr:`best_ntree_limit`.
    .. note::
        This parameter replaces `early_stopping_rounds` in :py:meth:`fit` method.
callbacks : Optional[List[TrainingCallback]]
    List of callback functions that are applied at end of each iteration.
    It is possible to use predefined callbacks by using
    :ref:`Callback API <callback_api>`.
    .. note::
       States in callback are not preserved during training, which means callback
       objects can not be reused for multiple training sessions without
       reinitialization or deepcopy.
    .. code-block:: python
        for params in parameters_grid:
            # be sure to (re)initialize the callbacks before each run
            callbacks = [xgb.callback.LearningRateScheduler(custom_rates)]
            xgboost.train(params, Xy, callbacks=callbacks)
kwargs : dict, optional
    Keyword arguments for XGBoost Booster object.  Full documentation of parameters
    can be found :doc:`here </parameter>`.
    Attempting to set a parameter via the constructor args and \\*\\*kwargs
    dict simultaneously will result in a TypeError.
    .. note:: \\*\\*kwargs unsupported by scikit-learn
        \\*\\*kwargs is unsupported by scikit-learn.  We do not guarantee
        that parameters passed via this argument will interact properly
        with scikit-learn.
"""


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class XGBRegressorConfig(ModelConfig):
    n_estimators: int = 100
    max_depth: int | None = None
    max_leaves: int | None = None
    max_bin: int | None = None
    grow_policy: int | None = None
    learning_rate: float | None = None
    verbosity: int | None = None
    objective: _SklObjective = None
    booster: str | None = None
    tree_method: str | None = None
    n_jobs: int | None = None
    gamma: float | None = None
    min_child_weight: float | None = None
    max_delta_step: float | None = None
    subsample: float | None = None
    sampling_method: str | None = None
    colsample_bytree: float | None = None
    colsample_bylevel: float | None = None
    colsample_bynode: float | None = None
    reg_alpha: float | None = None
    reg_lambda: float | None = None
    scale_pos_weight: float | None = None
    base_score: float | None = None
    random_state: int | RandomState | None = None
    missing: float | None = None
    num_parallel_tree: int | None = None
    monotone_constraints: dict[str, int] | str | None = None
    interaction_constraints: str | list[tuple[str]] | None = None
    importance_type: str | None = None
    gpu_id: int | None = None
    validate_parameters: bool | None = None
    predictor: str | None = None
    enable_categorical: bool = False
    max_cat_to_onehot: int | None = None
    eval_metric: str | list[str] | Callable | None = None
    early_stopping_rounds: int | None = None
    callbacks: list[TrainingCallback] | None = None
    kwargs: dict | None = None


class XGBoostRegressor(SKLearnModelBase[XGBRegressorConfig]):
    @property
    def _config_builder(self) -> type[XGBRegressorConfig]:
        return XGBRegressorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> XGBR:
        config = self.model_config

        # Converting None to np.nan, as required by XGBoost
        missing = np.nan if not config.missing else config.missing

        return XGBR(
            n_estimators=config.n_estimators,
            max_depth=config.max_depth,
            max_leaves=config.max_leaves,
            max_bin=config.max_bin,
            grow_policy=config.grow_policy,
            learning_rate=config.learning_rate,
            verbosity=config.verbosity,
            objective=config.objective,
            booster=config.booster,
            tree_method=config.tree_method,
            n_jobs=config.n_jobs,
            gamma=config.gamma,
            min_child_weight=config.min_child_weight,
            max_delta_step=config.max_delta_step,
            subsample=config.subsample,
            sampling_method=config.sampling_method,
            colsample_bytree=config.colsample_bytree,
            colsample_bylevel=config.colsample_bylevel,
            colsample_bynode=config.colsample_bynode,
            reg_alpha=config.reg_alpha,
            reg_lambda=config.reg_lambda,
            scale_pos_weight=config.scale_pos_weight,
            base_score=config.base_score,
            random_state=config.random_state,
            missing=missing,
            num_parallel_tree=config.num_parallel_tree,
            monotone_constraints=config.monotone_constraints,
            interaction_constraints=config.interaction_constraints,
            importance_type=config.importance_type,
            gpu_id=config.gpu_id,
            validate_parameters=config.validate_parameters,
            predictor=config.predictor,
            enable_categorical=config.enable_categorical,
            max_cat_to_onehot=config.max_cat_to_onehot,
            eval_metric=config.eval_metric,
            early_stopping_rounds=config.early_stopping_rounds,
            callbacks=config.callbacks,
            kwargs=config.kwargs,
        )



"""Maximum residual error."""

import logging
from typing import Any

import evaluate
from sklearn.metrics import max_error

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
This metric computes the maximum residual error, a metric that captures the
worst case error between the predicted value and the true value. In a perfectly
fitted single output regression model, the maximum residual error would be 0 on
the training set and though this would be highly unlikely in the real world,
this metric shows the extent of error that the model had when it was fitted.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated target values.
    references: Ground truth (correct) target values.

Returns:
    max_error: A positive floating point value (the best value is 0.0).

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("max_error")
    >>> predictions = [9, 2, 7, 1]
    >>> references = [3, 2, 7, 1]
    >>> metric.compute(predictions=predictions, references=references)
    {'max_error': 6.0}
"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class MaxError(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("float"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.max_error.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        **kwargs: Any,
    ) -> MetricResult:
        score = max_error(y_true=references, y_pred=predictions)
        return {self.tag: score}



from molflux.features.errors import FeaturisationError


def test_featurisation_error_is_runtime_error():
    """That our custom FeaturisationErrors inherit from RuntimeError.

    This helps narrow the scope of the exception, and allow to except it
    even in libraries not directly importing the package.
    """
    assert issubclass(FeaturisationError, RuntimeError)



import logging
from collections.abc import Iterator
from typing import Any, Protocol, runtime_checkable

from molflux.metrics.errors import DuplicateKeyError
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)


@runtime_checkable
class Metric(Protocol):
    """The public protocol for Metrics."""

    def __init__(self, **kwargs: Any) -> None:
        """Initialises the metric."""

    def __len__(self) -> int:
        """The number of examples loaded in the metric's cache."""

    @property
    def metadata(self) -> dict[str, Any]:
        """Dictionary containing all the metadata."""

    @property
    def name(self) -> str:
        """The canonical name."""

    @property
    def tag(self) -> str:
        """The tag name."""

    @property
    def state(self) -> dict[str, Any]:
        """The internal state."""

    def add_batch(
        self,
        *,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
        **kwargs: Any,
    ) -> None:
        """Add a batch of predictions and references for the metric's stack."""

    def compute(
        self,
        *,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
        **kwargs: Any,
    ) -> MetricResult:
        """Computes the metric."""

    def reset_state(self) -> None:
        """Resets the state."""

    def update_state(self, **kwargs: Any) -> None:
        """Pre-configures keyword arguments for compute()."""


class Metrics:
    """A collection of Metrics."""

    def __init__(self) -> None:
        self._stack: dict[str, Metric] = {}

    def __contains__(self, item: str) -> bool:
        return item in self._stack

    def __getitem__(self, key: str) -> Metric:
        metric = self._stack.get(key)
        if metric is None:
            raise KeyError(
                f"Metric {key!r} has not been loaded. Available metrics: {list(self._stack.keys())!r}",
            )
        return metric

    def __iter__(self) -> Iterator[Metric]:
        """Iterates through Metrics as if a list."""
        return iter(self._stack.values())

    def __len__(self) -> int:
        return len(self._stack)

    def __repr__(self) -> str:
        tags = sorted(metric.tag for metric in self._stack.values())
        return f"Metrics({tags!r})"

    def __setitem__(self, key: str, value: Metric) -> None:
        if key in self._stack:
            raise DuplicateKeyError(
                f"Metric with key {key!r} has already been added. You can add a unique custom tag to one of the metrics to add it to the collection.",
            )
        self._stack[key] = value

    def add_batch(
        self,
        *,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
        **kwargs: Any,
    ) -> None:
        for metric in self._stack.values():
            metric.add_batch(predictions=predictions, references=references, **kwargs)

    def add_metric(self, metric: Metric) -> None:
        self[metric.tag] = metric

    def compute(
        self,
        *,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
        **kwargs: Any,
    ) -> MetricResult:
        results = (
            metric.compute(
                predictions=predictions,
                references=references,
                **kwargs,
            )
            for metric in self._stack.values()
        )
        merged_results = {k: v for r in results for k, v in r.items()}
        return merged_results



import functools
import importlib.metadata

from cloudpathlib import AnyPath

from molflux.core.typing import PathLike

_REQUIREMENTS_FILENAME = "requirements.txt"


def pip_working_set() -> dict[str, str]:
    return _cached_pip_working_set().copy()


@functools.lru_cache
def _cached_pip_working_set() -> dict[str, str]:
    distributions = importlib.metadata.distributions()
    return {
        distribution.metadata["Name"]: distribution.version
        for distribution in distributions
    }


@functools.lru_cache
def _get_pinned_pip_environment() -> list[str]:
    """Returns a list of pinned package versions in the active python environment.

    This is cached for performance reasons.
    """
    return sorted(
        [f"{package}=={version}" for package, version in pip_working_set().items()],
    )


def save_python_environment(path: PathLike) -> str:
    """Saves metadata about the active python environment to disk."""
    requirements = _get_pinned_pip_environment()
    output_file = f"{path}/{_REQUIREMENTS_FILENAME}"
    with AnyPath(output_file).open("w") as f:  # type: ignore[attr-defined]
        f.write("\n".join(requirements))
    return output_file



import importlib.metadata
from dataclasses import asdict, dataclass, field
from typing import Any


def get_runtime() -> dict[str, str]:
    distributions = importlib.metadata.distributions()
    return {
        distribution.metadata["Name"]: distribution.version
        for distribution in distributions
    }


@dataclass
class ModelInfo:
    """
    Information about a model.

    See the constructor arguments and properties for a full list.

    Note: Not all fields are known on construction and may be updated later.
    """

    # Model class identifiers
    name: str = field(default_factory=str)

    # Generic model metadata
    tag: str = field(default_factory=str)
    version: str | None = None

    model_description: str = field(default_factory=str)

    task_types: list[str] = field(default_factory=list)

    # Model config metadata
    config: dict[str, Any] = field(default_factory=dict)
    config_description: str = field(default_factory=str)

    # Runtime metadata
    runtime: dict[str, str] = field(default_factory=get_runtime)

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)

    @classmethod
    def from_dict(cls, dictionary: dict[str, Any]) -> "ModelInfo":
        return cls(**dictionary)



import pytest

from molflux.features import Representation, list_representations, load_representation
from molflux.features.representations.rdkit.reaction.drfp import DRFP

representation_name = "drfp"


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, DRFP)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)


def test_default_representation_tag_matches_entrypoint_name(fixture_representation):
    """That the default representation tag matches the entrypoint name.

    This is not strictly required, but ensures a more intuitive user experience.
    """
    representation = fixture_representation
    assert representation.tag == representation_name


@pytest.mark.parametrize(
    "samples,expected_result",
    [
        ([], []),
        (
            [
                "CO.O.O=C(NC(=S)Nc1nc(-c2ccccc2)cs1)c1ccccc1.[Na+].[OH-]>>NC(=S)Nc1nc(-c2ccccc2)cs1",
            ],
            [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1]],
        ),
        (
            [
                "CO.O.O=C(NC(=S)Nc1nc(-c2ccccc2)cs1)c1ccccc1.[Na+].[OH-]>>NC(=S)Nc1nc(-c2ccccc2)cs1",
                "CO.O.O=C(NC(=S)Nc1nc(-c2ccccc2)cs1)c1ccccc1.[Na+].[OH-]>>NC(=S)Nc1nc(-c2ccccc2)cs1",
            ],
            [
                [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1],
                [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1],
            ],
        ),
    ],
)
def test_default_compute(fixture_representation, samples, expected_result):
    """That default scoring gives expected results."""
    representation = fixture_representation
    result = representation.featurise(samples, length=16)
    assert "drfp" in result
    assert result["drfp"] == expected_result



import datasets
from molflux.datasets.splitting import split_dataset


def test_returns_dataset_dict(fixture_splitting_strategy_mock, fixture_dataset):
    """That splitting returns a DatasetDict"""
    strategy = fixture_splitting_strategy_mock
    dataset = fixture_dataset
    splits = split_dataset(dataset=dataset, strategy=strategy)
    for split in splits:
        assert isinstance(split, datasets.DatasetDict)


def test_splits_are_datasets(fixture_splitting_strategy_mock, fixture_dataset):
    """That each returned split is a Dataset."""
    strategy = fixture_splitting_strategy_mock
    dataset = fixture_dataset
    splits = split_dataset(dataset=dataset, strategy=strategy)
    for split in splits:
        for dataset in split.values():
            assert isinstance(dataset, datasets.Dataset)


def test_generates_train_split(fixture_splitting_strategy_mock, fixture_dataset):
    """That splitting generates a train dataset."""
    strategy = fixture_splitting_strategy_mock
    dataset = fixture_dataset
    splits = split_dataset(dataset=dataset, strategy=strategy)
    for split in splits:
        assert "train" in split


def test_generates_validation_split(fixture_splitting_strategy_mock, fixture_dataset):
    """That splitting generates a validation dataset."""
    strategy = fixture_splitting_strategy_mock
    dataset = fixture_dataset
    splits = split_dataset(dataset=dataset, strategy=strategy)
    for split in splits:
        assert "validation" in split


def test_generates_test_split(fixture_splitting_strategy_mock, fixture_dataset):
    """That splitting generates a test dataset."""
    strategy = fixture_splitting_strategy_mock
    dataset = fixture_dataset
    splits = split_dataset(dataset=dataset, strategy=strategy)
    for split in splits:
        assert "test" in split



import functools
from importlib.metadata import entry_points

import pytest

from molflux.modelzoo.catalogue import NAMESPACE
from molflux.modelzoo.errors import ExtrasDependencyImportError
from molflux.modelzoo.load import load_model


@functools.lru_cache
def all_modelzoo_entrypoints() -> dict[str, list[str]]:
    """Cached for performance reasons."""
    return {
        namespace: [entrypoint.name for entrypoint in entrypoints]
        for namespace, entrypoints in entry_points().items()
        if namespace.startswith(NAMESPACE)
    }


_CATBOOST_MODELS = all_modelzoo_entrypoints()["molflux.modelzoo.plugins.catboost"]


@pytest.mark.parametrize("model_name", _CATBOOST_MODELS)
def test_missing_catboost_dependencies_raise_on_load(model_name):
    """That if dependencies are missing for catboost models, an
    error is raised at load time.

    The error should hint at how to install the missing dependencies.
    """

    with pytest.raises(
        ExtrasDependencyImportError,
        match=r"pip install \'molflux\[catboost\]\'",
    ):
        load_model(model_name)


_PYOD_MODELS = all_modelzoo_entrypoints()["molflux.modelzoo.plugins.pyod"]


@pytest.mark.parametrize("model_name", _PYOD_MODELS)
def test_missing_pyod_dependencies_raise_on_load(model_name):
    """That if dependencies are missing for pyod models, an
    error is raised at load time.

    The error should hint at how to install the missing dependencies.
    """

    with pytest.raises(
        ExtrasDependencyImportError,
        match=r"pip install \'molflux\[pyod\]\'",
    ):
        load_model(model_name)


_XGBOOST_MODELS = all_modelzoo_entrypoints()["molflux.modelzoo.plugins.xgboost"]


@pytest.mark.parametrize("model_name", _XGBOOST_MODELS)
def test_missing_xgboost_dependencies_raise_on_load(model_name):
    """That if dependencies are missing for xgboost models, an
    error is raised at load time.

    The error should hint at how to install the missing dependencies.
    """

    with pytest.raises(
        ExtrasDependencyImportError,
        match=r"pip install \'molflux\[xgboost\]\'",
    ):
        load_model(model_name)



import functools
import inspect
import logging
from collections import defaultdict
from collections.abc import Callable
from importlib.metadata import EntryPoint, entry_points
from types import ModuleType

import datasets.builder
from molflux.datasets.naming import camelcase_to_snakecase

logger = logging.getLogger(__name__)

NAMESPACE = "molflux.datasets.plugins."

# This is where entrypoints will be registered {<name>: <Entrypoint>}
BUILDERS_CATALOGUE: dict[str, EntryPoint] = {}


@functools.lru_cache
def fill_catalogue() -> None:
    """Fills the catalogue with entrypoints.

    Only ever call this once for performance reasons.
    """
    entrypoints = [
        entrypoint
        for namespace, entrypoints in entry_points().items()
        if namespace.startswith(NAMESPACE)
        for entrypoint in entrypoints
    ]
    for entrypoint in entrypoints:
        put_builder_entrypoint(entrypoint=entrypoint)


def put_builder_entrypoint(entrypoint: EntryPoint) -> None:
    """Puts a dataset builder entrypoint in the catalogue."""
    builder_name = camelcase_to_snakecase(entrypoint.name)

    if builder_name in BUILDERS_CATALOGUE:
        if entrypoint == BUILDERS_CATALOGUE[builder_name]:
            pass
        else:
            raise KeyError(f"Duplicate dataset builder: {builder_name!r}")

    BUILDERS_CATALOGUE[builder_name] = entrypoint


def get_builder_entrypoint(name: str) -> EntryPoint:
    """Returns a specific dataset builder entrypoint from the catalogue"""
    entrypoint = BUILDERS_CATALOGUE.get(name)
    if entrypoint is None:
        raise NotImplementedError(f"Dataset builder {name!r} is not available.")
    return entrypoint


def get_dataset_builder_path(name: str) -> str:
    entrypoint = get_builder_entrypoint(name=name)

    # Might raise (e.g. missing dependency extras)
    module = entrypoint.load()

    if not isinstance(module, ModuleType):
        raise NotImplementedError(f"Plugin {entrypoint.value} is not a valid module.")

    path = module.__file__

    if path is None:
        raise RuntimeError(
            f"Could not resolve path for plugin module {entrypoint.value}",
        )

    return path


def list_datasets() -> dict[str, list[str]]:
    """List all available datasets in the catalogue.

    The catalogue is returned as a view of dataset names keyed by kind.
    """

    view: dict[str, list[str]] = defaultdict(list)

    for name, entrypoint in BUILDERS_CATALOGUE.items():
        kind = entrypoint.group.split(".")[-1]
        view[kind].append(name)

    return dict(sorted(view.items()))


def register_builder(kind: str, name: str) -> Callable:
    """Registers a custom dataset builder in the entrypoints catalogue.

    Examples:
        .. code-block:: python

           @register_dataset(kind="custom", name="my_dataset")
           class CustomDataset:
               ...
    """

    def wrapper(model_cls: datasets.DatasetBuilder) -> datasets.DatasetBuilder:
        module = inspect.getmodule(model_cls).__name__  # type: ignore
        value = f"{module}"
        group = NAMESPACE + kind
        entrypoint = EntryPoint(name=name, value=value, group=group)
        put_builder_entrypoint(entrypoint=entrypoint)
        return model_cls

    return wrapper



from collections.abc import Iterable, Iterator, Sized
from typing import Any

import pytest

from molflux.datasets import load_dataset
from molflux.datasets.interfaces import SplittingStrategy


@pytest.fixture(scope="module")
def fixture_dataset(fixture_path_to_assets):
    """A sample Dataset implementation."""
    return load_dataset("esol").select(range(100))


class SplittingStrategyMock:
    """A mock SplittingStrategy"""

    def split(
        self,
        dataset: Sized,
        y: Iterable | None = None,
        groups: Iterable | None = None,
        **kwargs: Any,
    ) -> Iterator[tuple[Iterable[int], Iterable[int], Iterable[int]]]:
        """Splits the dataset into thirds."""
        n = len(dataset)
        indices = range(n)
        yield indices[: n // 3], indices[n // 3 : 2 * n // 3], indices[2 * n // 3 :]


@pytest.fixture(scope="module")
def fixture_splitting_strategy_mock():
    strategy = SplittingStrategyMock()
    assert isinstance(strategy, SplittingStrategy)
    return strategy



import warnings
from collections.abc import Iterable
from typing import Any, Literal, overload

import datasets
from datasets import (
    Dataset,
    DatasetBuilder,
    DatasetDict,
    IterableDataset,
    IterableDatasetDict,
)
from molflux.datasets.catalogue import BUILDERS_CATALOGUE, get_dataset_builder_path
from molflux.datasets.parsers import Spec, dict_parser, yaml_parser
from molflux.datasets.typing import PathLike


def load_dataset_builder(
    name: str,
    config_name: str | None = None,
    **kwargs: Any,
) -> DatasetBuilder:
    """Load a dataset builder from the catalogue or from the hf hub.

    A dataset builder can be used to inspect general information that is
    required to build a dataset (cache directory, config, dataset info, etc.)
    without downloading the dataset itself.

    You can find the list of available datasets with :func:`molflux.datasets.list_datasets()`.

    .. seealso::

        This function is a light wrapper around :func:`datasets.load_dataset_builder()`

    Args:
        name: The name of the dataset builder to load. This maps to the 'path'
            argument of :func:`datasets.load_dataset_builder()`.
        config_name: The name of the dataset configuration to load.
            If `None`, the default configuration is loaded. This maps to the
            'name' argument of :func:`datasets.load_dataset_builder()`.
        kwargs: Any other keyword arguments accepted by :func:`datasets.load_dataset_builder()`.

    Returns:
        A dataset builder.
    """

    # Check first if this is a custom builder name, and resolve path
    is_custom_builder = name in BUILDERS_CATALOGUE
    if is_custom_builder:
        name = get_dataset_builder_path(name)
        # ensure Datasets is allowed to execute the loading script by default
        kwargs = {"trust_remote_code": True, **kwargs}

    with warnings.catch_warnings():
        # for our custom builders, ignore warning from missing huggingface hub repo card metadata [MLOPS-1125]
        if is_custom_builder:
            warnings.filterwarnings(
                "ignore",
                message="Repo card metadata block was not found",
                category=UserWarning,
                module="huggingface_hub.repocard",
            )

        hf_builder = datasets.load_dataset_builder(
            path=name,
            name=config_name,
            **kwargs,
        )
        return hf_builder


@overload
def load_dataset(
    name: str,
    config_name: str | None = None,
    *,
    split: str = "all",
    streaming: Literal[False] = False,
    **kwargs: Any,
) -> Dataset: ...


@overload
def load_dataset(
    name: str,
    config_name: str | None = None,
    *,
    split: str = "all",
    streaming: Literal[True],
    **kwargs: Any,
) -> IterableDataset: ...


@overload
def load_dataset(
    name: str,
    config_name: str | None = None,
    *,
    split: None,
    streaming: Literal[False] = False,
    **kwargs: Any,
) -> DatasetDict: ...


@overload
def load_dataset(
    name: str,
    config_name: str | None = None,
    *,
    split: None,
    streaming: Literal[True],
    **kwargs: Any,
) -> IterableDatasetDict: ...


def load_dataset(
    name: str,
    config_name: str | None = None,
    *,
    split: str | None = "all",
    streaming: bool = False,
    **kwargs: Any,
) -> Dataset | DatasetDict | IterableDataset | IterableDatasetDict:
    """Load a dataset (or dataset dict) from the catalogue or from the hf hub.

    You can find the list of available datasets with :func:`molflux.datasets.list_datasets()`.

    .. seealso::

        This function is a light wrapper around :func:`datasets.load_dataset()`

    Args:
        name: The name of the dataset to load. This maps to the 'path'
            argument of :func:`datasets.load_dataset()`.
        config_name: The name of the dataset configuration to load.
            If `None`, the default configuration is loaded. This maps to the
            'name' argument of :func:`datasets.load_dataset()`.
        split: The split of the data to load.
            If None, will return a `dict` with all available splits.
            If given, will return a single Dataset. Defaults to 'all' which
            will return a single dataset combining data from all splits.
        streaming: Whether to stream the dataset. If set to True, the
            data files are not download but it is streamed progressively while
            iterating on the dataset. An IterableDataset or IterableDatasetDict
            is returned instead in this case. Defaults to False.
        kwargs: Any other keyword arguments accepted by :func:`datasets.load_dataset()`.

    Returns:
        The dataset requested.
    """

    # Check first if this is a custom builder name, and resolve path to loading script
    is_custom_builder = name in BUILDERS_CATALOGUE
    if is_custom_builder:
        name = get_dataset_builder_path(name)
        # ensure Datasets is allowed to execute the loading script by default
        kwargs = {"trust_remote_code": True, **kwargs}

    with warnings.catch_warnings():
        # for our custom builders, ignore warning from missing huggingface hub repo card metadata
        if is_custom_builder:
            warnings.filterwarnings(
                "ignore",
                message="Repo card metadata block was not found",
                category=UserWarning,
                module="huggingface_hub.repocard",
            )

        # Do not pass through unhandled deprecation warnings from dependencies of hf datasets themselves
        warnings.filterwarnings(
            "ignore",
            message="The .* keyword in pd.read_csv is deprecated and will be removed in a future version",
            category=FutureWarning,
            module="datasets[.*]",
        )
        return datasets.load_dataset(
            path=name,
            name=config_name,
            split=split,
            streaming=streaming,
            **kwargs,
        )


def _load_from_spec(spec: Spec) -> Dataset | DatasetDict:
    """Loads a dataset from a validated Spec."""

    dataset: Dataset | DatasetDict = load_dataset(name=spec.name, **spec.config)

    return dataset


def load_from_dict(dictionary: dict[str, Any]) -> Dataset | DatasetDict:
    """Loads a dataset from a config dict."""

    # Validate dictionary
    spec = dict_parser(dictionary=dictionary)

    return _load_from_spec(spec=spec)


Datasets = dict[str, Dataset | DatasetDict | IterableDataset | IterableDatasetDict]


def load_from_dicts(dictionaries: Iterable[dict[str, Any]]) -> Datasets:
    """Loads a collection of datasets from an iterable of dicts."""

    datasets = (load_from_dict(dictionary) for dictionary in dictionaries)

    # TODO: implement tags for datasets
    return {f"dataset-{i}": dataset for i, dataset in enumerate(datasets)}


def load_from_yaml(path: PathLike) -> Datasets:
    """Loads a collection of dataset from a yaml config file."""

    specs = yaml_parser(path=path)

    datasets = (_load_from_spec(spec) for spec in specs)

    # TODO: implement tags for datasets
    return {f"dataset-{i}": dataset for i, dataset in enumerate(datasets)}



"""
Tests ensuring desired API objects are part of the top-level namespace.
"""

import molflux.modelzoo


def test_exports_classification_mixin():
    """That the package exposes the ClassificationMixin ABC.

    This is exposed as a convenience building block for developers wishing to
    quickly create new modelzoo models in their applications.
    """
    assert hasattr(molflux.modelzoo, "ClassificationMixin")


def test_exports_list_models():
    """That the package exposes the list_models function."""
    assert hasattr(molflux.modelzoo, "list_models")
    assert callable(molflux.modelzoo.list_models)


def test_exports_load_model():
    """That the package exposes the load_model function."""
    assert hasattr(molflux.modelzoo, "load_model")
    assert callable(molflux.modelzoo.load_model)


def test_exports_load_from_dict():
    """That the package exposes the load_from_dict function."""
    assert hasattr(molflux.modelzoo, "load_from_dict")


def test_exports_load_from_dicts():
    """That the package exposes the load_from_dicts function."""
    assert hasattr(molflux.modelzoo, "load_from_dicts")


def test_exports_load_from_yaml():
    """That the package exposes the load_from_yaml function."""
    assert hasattr(molflux.modelzoo, "load_from_yaml")


def test_exports_register_model():
    """That the package exposes the register_model function."""
    assert hasattr(molflux.modelzoo, "register_model")
    assert callable(molflux.modelzoo.register_model)


def test_exports_model():
    """That the package exposes the Model type alias."""
    assert hasattr(molflux.modelzoo, "Model")


def test_exports_models():
    """That the package exposes the Models type alias."""
    assert hasattr(molflux.modelzoo, "Models")


def test_exports_model_base():
    """That the package exposes the ModelBase ABC.

    This is exposed as a convenience building block for developers wishing to
    quickly create new modelzoo models in their applications.
    """
    assert hasattr(molflux.modelzoo, "ModelBase")


def test_exports_supports_classification():
    """That the package exposes the supports_classification function."""
    assert hasattr(molflux.modelzoo, "supports_classification")
    assert callable(molflux.modelzoo.supports_classification)


def test_exports_supports_covariance():
    """That the package exposes the supports_covariance function."""
    assert hasattr(molflux.modelzoo, "supports_covariance")
    assert callable(molflux.modelzoo.supports_covariance)


def test_exports_supports_supports_prediction_interval():
    """That the package exposes the supports_prediction_interval function."""
    assert hasattr(molflux.modelzoo, "supports_prediction_interval")
    assert callable(molflux.modelzoo.supports_prediction_interval)


def test_exports_supports_sampling():
    """That the package exposes the supports_sampling function."""
    assert hasattr(molflux.modelzoo, "supports_sampling")
    assert callable(molflux.modelzoo.supports_sampling)


def test_exports_supports_supports_std():
    """That the package exposes the supports_std function."""
    assert hasattr(molflux.modelzoo, "supports_std")
    assert callable(molflux.modelzoo.supports_std)


def test_exports_supports_uncertainty_calibration():
    """That the package exposes the supports_uncertainty_calibration function."""
    assert hasattr(molflux.modelzoo, "supports_uncertainty_calibration")
    assert callable(molflux.modelzoo.supports_uncertainty_calibration)



from typing import Any

try:
    from rdkit.Chem.rdmolops import LayeredFingerprint
    from rdkit.DataStructs.cDataStructs import ExplicitBitVect

    from molflux.features.representations.rdkit._utils import rdkit_mol_from_smiles
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
Layered fingerprint for a molecule.

A topological (Daylight like) fingerprint for a molecule using a layer-based
hashing algorithm. These are another “RDKit original” and were developed with
the intention of using them as a substructure fingerprint. Since the pattern
fingerprint is far simpler and has proven to be quite effective as a
substructure fingerprint, the layered fingerprint hasn`t received much
attention. It may still be interesting for something, so we continue to include
it.

The idea of the fingerprint is generate features using the same subgraph
(or path) enumeration algorithm used in the RDKit fingerprint.
After a subgraph has been generated, it is used to set multiple bits based on
different atom and bond type definitions.
"""


class Layered(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        layer_flags: int = 0xFFFFFFFF,
        min_path: int = 1,
        max_path: int = 7,
        fp_size: int = 2048,
        atom_counts: list[int] | None = None,
        set_only_bits: ExplicitBitVect | None = None,
        branched_paths: bool = True,
        from_atoms: list | None = None,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """Generates a layered fingerprint for each input molecule.

        Args:
            samples: The molecules to be fingerprinted.
            layer_flags: The layers to be included (see below)
            min_path: The minimum path length (in bonds) to be included.
                Defaults to `1`.
            max_path: The minimum path length (in bonds) to be included.
                Defaults to `7`.
            fp_size: The size of the fingerprint. Defaults to `2048`.
            atom_counts: If not `None`, this will be used to provide the count
                of the number of paths that set bits each atom is involved in.
                The vector should have at least as many entries as the molecule
                has atoms and is not zeroed out here. Defaults to `None`.
            set_only_bits: If provided, only bits that are set in this bit
                vector will be set in the result. This is essentially the same
                as doing: `(*res) &= (*setOnlyBits)`; but also has an impact on
                the `atom_counts` (if being used).
            branched_paths: Toggles generation of branched subgraphs, not just
                linear paths. Defaults to `None`.
            from_atoms: If provided, only the atoms in the vector will be used
                as centers in the fingerprint. Defaults to `None`.

        Layer definitions:

            * 0x01: pure topology
            * 0x02: bond order
            * 0x04: atom types
            * 0x08: presence of rings
            * 0x10: ring sizes
            * 0x20: aromaticity

        Returns:
            Layered fingerprints, as lists of bits.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('layered')
            >>> samples = ['c1ccccc1']
            >>> representation.featurise(samples, fp_size=16)
            {'layered': [[1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        layered_fp_list: list[list] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smile = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smile)
                rd_fp = LayeredFingerprint(
                    mol,
                    layerFlags=layer_flags,
                    minPath=min_path,
                    maxPath=max_path,
                    fpSize=fp_size,
                    atomCounts=atom_counts or [],
                    setOnlyBits=set_only_bits,
                    branchedPaths=branched_paths,
                    fromAtoms=from_atoms or [],
                )

                if rd_fp.GetNumOnBits() == 0:
                    layered_fp_list.append([0] * rd_fp.GetNumBits())
                else:
                    layered_fp_list.append(rd_fp.ToList())

        return {self.tag: layered_fp_list}



from typing import Any

import pytest

import datasets
from molflux.datasets import featurise_dataset


class MockRepresentation:
    """Implements the interfaces.Representation protocol.

    Here we define a single representation that simply return the input samples
    untouched. This can be used to simulate one-to-one representations.
    """

    def __init__(self, name: str) -> None:
        self.name = name

    def featurise(self, *columns, **kwargs):
        return {self.name: columns[0]}


class MockMultiRepresentation:
    """Implements the interfaces.Representation protocol.

    Here we define a single representation that returns n copies of the input
    samples untouched. This can be used to simulate one-to-many representations.
    """

    def __init__(self, name: str, n: int) -> None:
        self.name = name
        self.n = n

    def featurise(self, *columns: Any, **kwargs: Any) -> dict[str, Any]:
        samples = columns[0]
        return {f"{self.name}{i}": samples for i in range(self.n)}


class MockMultiColumnRepresentation:
    """Implements the interfaces.Representation protocol.

    Here we define a single representation that sums the input columns.
    This can be used to simulate many-to-one representations.
    """

    def __init__(self, name: str) -> None:
        self.name = name

    def featurise(self, *columns: Any, **kwargs: Any) -> dict[str, Any]:
        concat_columns = [sum(samples) for samples in zip(*columns, strict=False)]
        return {self.name: concat_columns}


@pytest.fixture(scope="module")
def fixture_mock_dataset() -> datasets.Dataset:
    data = {
        "a": [1, 2, 3, 4, 5],
        "b": [6, 7, 8, 9, 10],
    }
    dataset = datasets.Dataset.from_dict(data)
    return dataset


@pytest.fixture(scope="module")
def fixture_mock_dataset_dict() -> datasets.DatasetDict:
    ds_train = datasets.Dataset.from_dict({"a": [1, 2, 3, 4]})
    ds_test = datasets.Dataset.from_dict({"a": [1, 3, 4, 2]})
    return datasets.DatasetDict({"train": ds_train, "test": ds_test})


@pytest.fixture(scope="module")
def fixture_mock_dataset_dict_with_empty_splits() -> datasets.DatasetDict:
    ds_train = datasets.Dataset.from_dict({"a": []})
    ds_validation = datasets.Dataset.from_dict({"a": [1, 3, 4, 2]})
    ds_test = datasets.Dataset.from_dict({"a": []})
    return datasets.DatasetDict(
        {"train": ds_train, "validation": ds_validation, "test": ds_test},
    )


@pytest.fixture(scope="module")
def fixture_mock_one_to_one_representation():
    """A single representation that returns one arbitrary output."""
    return MockRepresentation("original_name_a")


@pytest.fixture(scope="module")
def fixture_mock_one_to_many_representation():
    """A single representation  that returns 3 arbitrary outputs."""
    return MockMultiRepresentation("original_name_a", n=3)


@pytest.fixture(scope="module")
def fixture_mock_many_to_one_representation():
    """A single representation taking arbitrarily many columns and returning
    one"""
    return MockMultiColumnRepresentation("original_name_a")


@pytest.fixture(scope="module")
def fixture_mock_representations():
    """A collection of representations that simulates an heterogeneous mix of
    representations including representations generating multiple features.
    """

    return [
        MockRepresentation("original_name_a"),
        MockMultiRepresentation("original_name_b", n=2),
        MockRepresentation("original_name_c"),
    ]


@pytest.mark.parametrize(
    "representations_fixture",
    [
        "fixture_mock_one_to_many_representation",
        "fixture_mock_representations",
    ],
)
def test_can_use_representation_or_representations(
    fixture_mock_dataset,
    representations_fixture,
    request,
):
    """That can featurise a dataset both providing a single representation (convenience)
    or a collection of representations (core)."""
    dataset = fixture_mock_dataset
    representations = request.getfixturevalue(representations_fixture)
    column_to_featurise = dataset.column_names[0]
    assert featurise_dataset(
        dataset=dataset,
        column=column_to_featurise,
        representations=representations,
    )


def test_featurise_dataset_raises_on_invalid_source_column(
    fixture_mock_dataset,
    fixture_mock_representations,
):
    """That an error is raised if attempting to featurise a non-existent column."""
    dataset = fixture_mock_dataset
    representations = fixture_mock_representations

    with pytest.raises(KeyError, match=r"Feature .+? not in dataset"):
        featurise_dataset(
            dataset=dataset,
            column="non-existent-column",
            representations=representations,
        )


def test_featurise_dataset_adds_features(
    fixture_mock_dataset,
    fixture_mock_representations,
):
    """That the featurised dataset has additional columns with new features."""
    dataset = fixture_mock_dataset
    representations = fixture_mock_representations
    column_to_featurise = dataset.column_names[0]
    featurised_dataset = featurise_dataset(
        dataset=dataset,
        column=column_to_featurise,
        representations=representations,
    )
    assert len(featurised_dataset.column_names) > len(dataset.column_names)


@pytest.mark.parametrize(
    ("display_names", "expected_output_names"),
    [
        (
            # canonical format (nested array with one array of display names for each representation)
            [["custom_name_a"]],
            ["custom_name_a"],
        ),
        (
            # shorthand for single representation (worth to support as we allow users to pass a single representation to the function)
            ["custom_name_a"],
            ["custom_name_a"],
        ),
        (
            # apply None (fallback) template across all representations
            None,
            ["a::original_name_a"],
        ),
        (
            # canonical format, apply fallback template across all outputs of associated representation (same effect in this case)
            [[None]],
            ["a::original_name_a"],
        ),
        (
            # same but using shorthand for single representation
            [None],
            ["a::original_name_a"],
        ),
        (
            # apply custom template across all representations
            "{source_column}>>{feature_name}",
            ["a>>original_name_a"],
        ),
        (
            # canonical format, apply custom template across all outputs of associated representation (same effect in this case)
            [["{source_column}>>{feature_name}"]],
            ["a>>original_name_a"],
        ),
        (
            # shorthand for single representation
            ["{source_column}>>{feature_name}"],
            ["a>>original_name_a"],
        ),
    ],
)
def test_allowed_display_names_notations_for_one_to_one_representation_outputs(
    fixture_mock_dataset,
    fixture_mock_one_to_one_representation,
    display_names,
    expected_output_names,
):
    """That can assign custom display names to the featurisation outputs obtained
    using a single one-to-one representation.

    All display names formats shown above should be accepted and give the
    corresponding output.
    """
    dataset = fixture_mock_dataset
    representation = fixture_mock_one_to_one_representation
    column_to_featurise = dataset.column_names[0]
    featurised_dataset = featurise_dataset(
        dataset=dataset,
        column=column_to_featurise,
        representations=representation,
        display_names=display_names,
    )
    for display_name in expected_output_names:
        assert display_name in featurised_dataset.column_names


@pytest.mark.parametrize(
    ("display_names", "expected_output_names"),
    [
        (
            # canonical format (nested array with one array of display names for each representation)
            [["custom_name_a0", "custom_name_a1", "custom_name_a2"]],
            ["custom_name_a0", "custom_name_a1", "custom_name_a2"],
        ),
        (
            # shorthand for single representation
            ["custom_name_a0", "custom_name_a1", "custom_name_a2"],
            ["custom_name_a0", "custom_name_a1", "custom_name_a2"],
        ),
        (
            # canonical format (1 fallback to default template)
            [["custom_name_a0", None, "custom_name_a2"]],
            ["custom_name_a0", "a::original_name_a1", "custom_name_a2"],
        ),
        (
            # canonical format (2 fallbacks to default template)
            [["custom_name_a0", None, None]],
            ["custom_name_a0", "a::original_name_a1", "a::original_name_a2"],
        ),
        (
            # canonical format (all fallbacks to default template, explicit)
            [[None, None, None]],
            ["a::original_name_a0", "a::original_name_a1", "a::original_name_a2"],
        ),
        (
            # canonical format (all fallbacks to default template, implicit)
            [[None]],
            ["a::original_name_a0", "a::original_name_a1", "a::original_name_a2"],
        ),
        (
            # shorthand for single representation
            [None],
            ["a::original_name_a0", "a::original_name_a1", "a::original_name_a2"],
        ),
        (
            # apply None (fallback) template across all representations (only one here)
            None,
            ["a::original_name_a0", "a::original_name_a1", "a::original_name_a2"],
        ),
        (
            # apply custom template across all representations
            "{source_column}>>{feature_name}",
            ["a>>original_name_a0", "a>>original_name_a1", "a>>original_name_a2"],
        ),
        # ...
    ],
)
def test_allowed_display_names_notations_for_one_to_many_representation_outputs(
    fixture_mock_dataset,
    fixture_mock_one_to_many_representation,
    display_names,
    expected_output_names,
):
    """That can assign custom display names to the featurisation outputs obtained
    using a single one-to-many representation.

    All display names formats shown above should be accepted and give the
    corresponding output.
    """
    dataset = fixture_mock_dataset
    representation = fixture_mock_one_to_many_representation
    column_to_featurise = dataset.column_names[0]
    featurised_dataset = featurise_dataset(
        dataset=dataset,
        column=column_to_featurise,
        representations=representation,
        display_names=display_names,
    )
    for display_name in expected_output_names:
        assert display_name in featurised_dataset.column_names


@pytest.mark.parametrize(
    ("pass_columns", "display_names", "expected_output_names"),
    [
        (
            # canonical format (nested array with one array of display names for each representation)
            ["a", "b"],
            [["custom_name_a"]],
            ["custom_name_a"],
        ),
        (
            # shorthand for single representation (worth to support as we allow users to pass a single representation to the function)
            ["a", "b"],
            ["custom_name_a"],
            ["custom_name_a"],
        ),
        (
            # apply None (fallback) template across all representations
            ["a", "b"],
            None,
            ["a:b::original_name_a"],
        ),
        (
            # canonical format, apply fallback template across all outputs of associated representation (same effect in this case)
            ["a", "b"],
            [[None]],
            ["a:b::original_name_a"],
        ),
        (
            # same but using shorthand for single representation
            ["a", "b"],
            [None],
            ["a:b::original_name_a"],
        ),
        (
            # apply custom template across all representations
            ["a", "b"],
            "{source_column}>>{feature_name}",
            ["a:b>>original_name_a"],
        ),
        (
            # canonical format, apply custom template across all outputs of associated representation (same effect in this case)
            ["a", "b"],
            [["{source_column}>>{feature_name}"]],
            ["a:b>>original_name_a"],
        ),
        (
            # shorthand for single representation
            ["a", "b"],
            ["{source_column}>>{feature_name}"],
            ["a:b>>original_name_a"],
        ),
        (
            # canonical format with single column
            ["a"],
            [["custom_name_a"]],
            ["custom_name_a"],
        ),
        (
            # canonical format, template and single column
            ["a"],
            [["{source_column}>>{feature_name}"]],
            ["a>>original_name_a"],
        ),
    ],
)
def test_allowed_display_names_notations_for_many_to_one_representation_outputs(
    fixture_mock_dataset,
    fixture_mock_many_to_one_representation,
    pass_columns,
    display_names,
    expected_output_names,
):
    """That can assign custom display names to the featurisation outputs obtained
    using a single many-to-one representation.

    All display names formats shown above should be accepted and give the
    corresponding output.
    """
    dataset = fixture_mock_dataset
    representation = fixture_mock_many_to_one_representation
    featurised_dataset = featurise_dataset(
        dataset=dataset,
        column=pass_columns,
        representations=representation,
        display_names=display_names,
    )
    for display_name in expected_output_names:
        assert display_name in featurised_dataset.column_names


@pytest.mark.parametrize(
    ("display_names", "expected_output_names"),
    [
        (
            # canonical format (nested array with one array of display names for each representation)
            [
                ["custom_name_a"],
                ["custom_name_b0", "custom_name_b1"],
                ["custom_name_c"],
            ],
            ["custom_name_a", "custom_name_b0", "custom_name_b1", "custom_name_c"],
        ),
        (
            # canonical syntax, mix of custom display names and fallback templates
            [[None], ["custom_name_b0", None], ["custom_name_c"]],
            [
                "a::original_name_a",
                "custom_name_b0",
                "a::original_name_b1",
                "custom_name_c",
            ],
        ),
        (
            # canonical format (all fallback templates, explicit)
            [[None], [None, None], [None]],
            [
                "a::original_name_a",
                "a::original_name_b0",
                "a::original_name_b1",
                "a::original_name_c",
            ],
        ),
        (
            # canonical format (all fallback templates, using implicit syntax for multi-output feature)
            [[None], [None], [None]],
            [
                "a::original_name_a",
                "a::original_name_b0",
                "a::original_name_b1",
                "a::original_name_c",
            ],
        ),
        (
            # apply None (fallback) template across all representations
            None,
            [
                "a::original_name_a",
                "a::original_name_b0",
                "a::original_name_b1",
                "a::original_name_c",
            ],
        ),
        (
            # apply custom template across all representations
            "{source_column}>>{feature_name}",
            [
                "a>>original_name_a",
                "a>>original_name_b0",
                "a>>original_name_b1",
                "a>>original_name_c",
            ],
        ),
        # ...
    ],
)
def test_allowed_display_names_notations_for_representations_outputs(
    fixture_mock_dataset,
    fixture_mock_representations,
    display_names,
    expected_output_names,
):
    """That can assign custom display names to the featurisation outputs obtained
    using a single one-to-many representation.

    All display names formats shown above should be accepted for collections of
    representations and give the corresponding output.
    """
    dataset = fixture_mock_dataset
    representations = fixture_mock_representations
    column_to_featurise = dataset.column_names[0]
    featurised_dataset = featurise_dataset(
        dataset=dataset,
        column=column_to_featurise,
        representations=representations,
        display_names=display_names,
    )
    for display_name in expected_output_names:
        assert display_name in featurised_dataset.column_names


def test_using_shorthand_display_names_syntax_with_representations_raises(
    fixture_mock_dataset,
    fixture_mock_representations,
):
    """That using display names in the single-representation shorthand notation
    but when providing multiple representations raises an error.
    """
    dataset = fixture_mock_dataset
    representations = fixture_mock_representations
    column_to_featurise: str = dataset.column_names[0]
    shorthand_display_names = [
        "custom_name_a",
        "custom_name_b0",
        "custom_name_b1",
        "custom_name_c",
    ]
    with pytest.raises(ValueError, match="can only be used for single representations"):
        featurise_dataset(
            dataset=dataset,
            column=column_to_featurise,
            representations=representations,
            display_names=shorthand_display_names,  # type: ignore[arg-type]
        )


def test_using_unsupported_display_names_syntax_raises(
    fixture_mock_dataset,
    fixture_mock_representations,
):
    """That using display names in an unsupported syntax format raises an error."""
    dataset = fixture_mock_dataset
    representations = fixture_mock_representations
    column_to_featurise = dataset.column_names[0]
    mixed_notation_display_names = [
        "custom_name_a",
        ["custom_name_b0", "custom_name_b1"],
        "custom_name_c",
    ]
    with pytest.raises(ValueError, match="do not follow the accepted syntax"):
        featurise_dataset(
            dataset=dataset,
            column=column_to_featurise,
            representations=representations,
            display_names=mixed_notation_display_names,  # type: ignore[arg-type]
        )


def test_providing_not_enough_display_names_for_number_of_representations_used_raises(
    fixture_mock_dataset,
    fixture_mock_representations,
):
    """That not providing enough display names for the number of representations
    being used raises an error."""
    dataset = fixture_mock_dataset
    representations = fixture_mock_representations
    column_to_featurise = dataset.column_names[0]
    display_names = [
        ["custom_name_a"],
        [None],
    ]  # missing display names for third representation
    with pytest.raises(RuntimeError, match="Error parsing display names"):
        featurise_dataset(
            dataset=dataset,
            column=column_to_featurise,
            representations=representations,
            display_names=display_names,  # type: ignore[arg-type]
        )


def test_using_non_template_as_template_raises(
    fixture_mock_dataset,
    fixture_mock_representations,
):
    """That safeguards are in place against using a static string as template to
    be broadcasted across all output features. This would implicitly rename
    all outputs to the same static name, resulting in a single output feature."""
    dataset = fixture_mock_dataset
    representations = fixture_mock_representations
    column_to_featurise = dataset.column_names[0]
    mixed_notation_display_names = [
        ["custom_name_a"],
        ["not_a_template"],
        ["custom_name_c"],
    ]
    with pytest.raises(ValueError, match="Only template display names can be used"):
        featurise_dataset(
            dataset=dataset,
            column=column_to_featurise,
            representations=representations,
            display_names=mixed_notation_display_names,  # type: ignore[arg-type]
        )


def test_using_a_template_with_unknown_fields_raises(
    fixture_mock_dataset,
    fixture_mock_representations,
):
    """That setting a template requiring fields which are not made available
    by the backend templating context raises an error.
    """
    dataset = fixture_mock_dataset
    representations = fixture_mock_representations
    column_to_featurise = dataset.column_names[0]
    template = "{source_column}::{unsupported_key}"

    with pytest.raises(
        KeyError,
        match="Could not template the target string: 'unsupported_key'",
    ):
        featurise_dataset(
            dataset=dataset,
            column=column_to_featurise,
            representations=representations,
            display_names=template,
        )


def test_setting_display_name_already_in_use_warns(
    fixture_mock_dataset,
    fixture_mock_representations,
):
    """That setting a display name that is the same as the name of an existent
    column in the dataset raises a warning.

    This should be allowed to enable more complex / dynamic config driven
    featurisation workflows.
    """
    dataset = fixture_mock_dataset
    representations = fixture_mock_representations
    column_to_featurise = dataset.column_names[0]
    existing_column_name: str = dataset.column_names[1]
    display_names = [[existing_column_name], [None], [None]]

    # Check that the warning mentions the column being overwritten
    with pytest.warns(UserWarning, match=dataset.column_names[1]):
        featurise_dataset(
            dataset=dataset,
            column=column_to_featurise,
            representations=representations,
            display_names=display_names,  # type: ignore[arg-type]
        )


@pytest.mark.parametrize(
    "dataset_dict_fixture",
    [
        "fixture_mock_dataset_dict",
        "fixture_mock_dataset_dict_with_empty_splits",
    ],
)
def test_can_featurise_dataset_dict(
    dataset_dict_fixture,
    fixture_mock_representations,
    request,
):
    """That can featurise a DatasetDict.

    We would generally advise featurising a single `Dataset` and then splitting
    it into a `DatasetDict`, but feturising a `DatasetDict` should technically
    be supported too.
    """
    dataset_dict = request.getfixturevalue(dataset_dict_fixture)
    representations = fixture_mock_representations

    column_to_featurise = dataset_dict.column_names["train"][0]
    featurised_dataset_dict = featurise_dataset(
        dataset=dataset_dict,
        column=column_to_featurise,
        representations=representations,
    )

    # check there is the same number of Datasets
    assert len(featurised_dataset_dict) == len(dataset_dict)

    # check that all Datasets have been featurised
    for split_name, featurised_datast in featurised_dataset_dict.items():
        assert len(featurised_datast.column_names) > len(
            dataset_dict[split_name].column_names,
        )


def test_raises_on_inconsistent_dataset_dict(fixture_mock_representations):
    """That an error is raised if feeding DatasetDicts made of Datasets that
    do not all have the same features.

    This is not strictly required, but featurising DatasetDicts of heterogeneous
    collections of Datasets is not part of any official workflow we would want to
    support. This makes sures we highlight this kind of inconsistencies as early
    as possible, preventing hard to debug side effects down the road.
    """
    ds_train = datasets.Dataset.from_dict({"a": [1, 2], "c": [3, 4]})
    ds_test = datasets.Dataset.from_dict({"a": [2, 1], "b": [4, 3]})
    dataset_dict = datasets.DatasetDict({"train": ds_train, "test": ds_test})

    representations = fixture_mock_representations
    column_to_featurise = dataset_dict.column_names["train"][0]

    with pytest.raises(ValueError, match=r".*Inconsistent input features.*"):
        featurise_dataset(
            dataset=dataset_dict,
            column=column_to_featurise,
            representations=representations,
        )



from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    try:
        import stan
    except ImportError as e:
        from molflux.modelzoo.errors import ExtrasDependencyImportError

        raise ExtrasDependencyImportError("pystan", e) from None


class StanWrapper:
    def __init__(self, filename: str) -> None:
        path_to_stan_code = Path(__file__).parent / filename
        self.model_code: str = path_to_stan_code.read_text()
        self.fit: stan.fit.Fit | None = None  # to store posterior samples



import pathlib

import _pytest.config
import pytest

PytestConfig = _pytest.config.Config


def path_to_file(pytestconfig: PytestConfig, *other: str) -> pathlib.Path:
    """Get an absolute path to the given target file."""
    return pytestconfig.rootpath.joinpath(*other)


@pytest.fixture(scope="module")
def fixture_path_to_assets(pytestconfig: PytestConfig) -> pathlib.Path:
    """Get an absolute path to the assets directory."""
    return path_to_file(pytestconfig, "tests", "metrics", "assets")



from typing import Any

from colorama import Fore, Style


class DuplicateKeyError(KeyError):
    """Raisable when a mapping (dictionary) key is being duplicated."""


class FeaturisationError(RuntimeError):
    """Raisable on internal errors during featurisation calls."""

    def __init__(self, sample: Any):
        msg = f"Error processing sample {sample!r}"
        super().__init__(msg)


class ExtrasDependencyImportError(Exception):
    """Raisable on ImportErrors due to missing package extras."""

    def __init__(self, extras_type: str, nested_error: Exception):
        message = (
            f"\n\n📦 {nested_error}\n\n"
            + f"{Style.DIM}# have you tried running the following?{Style.RESET_ALL}\n"
            + f"$ {Style.BRIGHT + Fore.GREEN}pip install 'molflux[{extras_type}]'{Style.RESET_ALL}"
        )
        super().__init__(message)


class InvalidNumberOfPositionalArgs(Exception):
    """Raisable when the expected and actual number of columns in a representation featurise call differ"""

    def __init__(self, expected_size: int, actual_size: int):
        msg = f"Error processing *columns within the featuriser - {expected_size} positional arguments expected, {actual_size} received"
        super().__init__(msg)



import logging
from functools import reduce

try:
    from openeye import oechem, oequacpac
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

logger = logging.getLogger(__name__)

_DEFAULT_TAUTOMER_TIMEOUTS = [0.01, 0.1, 1, 10, 100]

# setting log levels for oequacpac, as it can be very verbose
oequacpac.oechem.OEThrow.SetLevel(oechem.OEErrorLevel_Error)


def ensure_oemol_title(mol: oechem.OEMolBase, title: str) -> oechem.OEMolBase:
    if mol.GetTitle() == "":
        mol.SetTitle(title)
    return mol


def standardise_oemol(
    mol: oechem.OEMolBase,
    *,
    strip_salts: bool = False,
    set_neutral_ph: bool = False,
    reasonable_protomer: bool = False,
    reasonable_tautomer: bool = False,
    explicit_h: bool = False,
    remove_formal_charges: bool = False,
    perceive_chiral: bool = True,
    assign_hyb: bool = True,
    rekekulize: bool = False,
    remove_stereo: bool = False,
    clear_non_chiral_stereo: bool = False,
    remove_non_standard_stereo: bool = False,
    sd_title_tag: str = "",
    clear_sd_data: bool = False,
    gen_2d_coords: bool = False,
    tautomer_options: oequacpac.OETautomerOptions | None = None,
    tautomer_timeouts: list[float] | None = None,
) -> oechem.OEMolBase:
    """
    Returns a standardised OEMol object.
    """

    if remove_formal_charges and (reasonable_protomer or set_neutral_ph):
        raise ValueError("cannot set neutral pH and remove formal charges.")

    # create a copy of the input molecule for standardisation
    std_mol = mol.CreateCopy()

    # use a given SD tag as molecule title
    if sd_title_tag:
        if isinstance(std_mol, oechem.OEMCMolBase):
            title = oechem.OEGetSDData(std_mol.GetActive(), sd_title_tag)
        else:
            title = oechem.OEGetSDData(std_mol, sd_title_tag)

        std_mol.SetTitle(title)

    if clear_sd_data:
        if isinstance(std_mol, oechem.OEGraphMol):
            oechem.OEClearSDData(std_mol)
        else:
            for conf in std_mol.GetConfs():
                oechem.OEClearSDData(conf)

    if strip_salts:
        oechem.OETheFunctionFormerlyKnownAsStripSalts(std_mol)

    if explicit_h and not oechem.OEHasExplicitHydrogens(std_mol):
        oechem.OEAddExplicitHydrogens(std_mol)

    if not reasonable_tautomer and (tautomer_options or tautomer_timeouts):
        logger.warning(
            "Tautomer options were given without setting reasonable_tautomer=True. No tautomerisation will"
            " be done, which might not be the intended user behaviour.",
        )

    if reasonable_tautomer:
        pka_norm = False

        if tautomer_options is None:
            tautomer_options = {}
        tautomer_options = generate_tautomer_options(**tautomer_options)

        if tautomer_timeouts is None:
            tautomer_timeouts = _DEFAULT_TAUTOMER_TIMEOUTS.copy()

        # flag for whether we found a reasonable tautomer
        successful_tautomerisation = False

        # go through each timeout and try to generate reasonable tautomers
        for max_timeout in tautomer_timeouts:
            tautomer_options.SetMaxSearchTime(max_timeout)

            reas_tauts = list(
                oequacpac.OEGetReasonableTautomers(std_mol, tautomer_options, pka_norm),
            )

            # if no reasonable tautomers were generated, skip to the next max_timeout
            if len(reas_tauts) == 0:
                continue

            # OEGetReasonableTautomers returns a list of 1 or more (depending on
            # tautomerOptions set); the first one is returned
            if len(reas_tauts) > 1:
                logger.debug(
                    "Multiple reasonable tautomers returned for molecule -> taking top ranked",
                )

            std_mol = reas_tauts[0]
            successful_tautomerisation = True
            break

        # if the loop above finished with no reasonable tautomer found at any stage, raise an error
        if not successful_tautomerisation:
            raise RuntimeError(
                f"Could not generate a reasonable tautomer in the given time for {oechem.OEMolToSmiles(std_mol)}",
            )

    if set_neutral_ph:
        oequacpac.OESetNeutralpHModel(std_mol)

    if reasonable_protomer:
        oequacpac.OEGetReasonableProtomer(std_mol)

    # remove formal charges / OERemoveFormalCharge returns False for molecules
    # with unknown atom types
    if remove_formal_charges:
        if not oequacpac.OERemoveFormalCharge(std_mol):
            logger.error("failed to remove formal charges from %s.", mol.GetTitle())

    if assign_hyb and not std_mol.HasPerceived(oechem.OEPerceived_Hybridization):
        oechem.OEAssignHybridization(std_mol)

    if remove_stereo:
        std_mol = remove_mol_stereo(std_mol)

    elif clear_non_chiral_stereo:
        clear_non_chiral_mol_stereo(std_mol)

    if remove_non_standard_stereo:
        remove_non_standard_mol_stereo(std_mol)

    if perceive_chiral and not std_mol.HasPerceived(oechem.OEPerceived_Chiral):
        oechem.OEPerceiveChiral(std_mol)

    if rekekulize:
        rekekulize_mol(std_mol)

    if gen_2d_coords:
        oechem.OEGenerate2DCoordinates(std_mol)

    return std_mol


def remove_mol_stereo(mol: oechem.OEMolBase) -> oechem.OEMolBase:
    """
    Returns a copy of the input molecule `mol` with all stereochemistry
    information removed.
    """
    cmol = mol.CreateCopy()
    cmol.Clear()

    oechem.OEUncolorMol(
        cmol,
        mol,
        oechem.OEUncolorStrategy_RemoveAtomStereo
        | oechem.OEUncolorStrategy_RemoveBondStereo
        | oechem.OEUncolorStrategy_RemoveGroupStereo,
    )

    return cmol


def clear_non_chiral_mol_stereo(mol: oechem.OEMolBase) -> None:
    """
    Clears all false stereo information from non-chiral atoms and bonds.
    """
    clear_non_chiral_atom_stereo(mol)
    clear_non_chiral_bond_stereo(mol)


def clear_non_chiral_atom_stereo(mol: oechem.OEMolBase) -> None:
    """
    Removes false atom stereo information from non-chiral atoms.
    """
    pred = oechem.OEAndAtom(
        oechem.OEHasAtomStereoSpecified(),
        oechem.OENotAtom(oechem.OEIsChiralAtom()),
    )

    for atom in mol.GetAtoms(pred):
        atom.SetStereo(
            list(atom.GetAtoms()),
            oechem.OEAtomStereo_Tetrahedral,
            oechem.OEAtomStereo_Undefined,
        )


def clear_non_chiral_bond_stereo(mol: oechem.OEMolBase) -> None:
    """
    Removes false bond stereo information from non-chiral bonds.
    """
    pred = reduce(
        oechem.OEAndBond,
        (
            oechem.OEHasBondStereoSpecified(),
            oechem.OENotBond(oechem.OEIsChiralBond()),
            oechem.OEHasOrder(2),
        ),
    )

    for bond in mol.GetBonds(pred):
        bond.SetStereo(
            [bond.GetBgn(), bond.GetEnd()],
            oechem.OEBondStereo_CisTrans,
            oechem.OEBondStereo_Undefined,
        )


def remove_non_standard_mol_stereo(mol: oechem.OEMolBase) -> None:
    """
    OpenEye's definition of what is chiral includes tertiary nitrogens.
    This isn't really from a synthetic chemists standpoint - so this
    function removes chiral centres from these.
    """
    pred = oechem.OEAndAtom(
        oechem.OEHasAtomStereoSpecified(),
        oechem.OEAndAtom(oechem.OEHasAtomicNum(7), oechem.OEHasHvyDegree(3)),
    )

    for atom in mol.GetAtoms(pred):
        atom.SetStereo(
            list(atom.GetAtoms()),
            oechem.OEAtomStereo_Tetrahedral,
            oechem.OEAtomStereo_Undefined,
        )


def rekekulize_mol(
    mol: oechem.OEMolBase,
    aro_model: int = oechem.OEAroModel_OpenEye,
) -> None:
    """
    Reassigns the single-double bond distribution in aromatic rings. In rare instances,
    identical molecules can have a different distribution, which can cause problems with
    algorithms that hash fragments.
    """
    oechem.OEFindRingAtomsAndBonds(mol)
    oechem.OEAssignAromaticFlags(mol, aro_model)

    for bond in mol.GetBonds():
        if bond.IsAromatic():
            bond.SetIntType(5)
        else:
            bond.SetIntType(bond.GetOrder())

    oechem.OECanonicalOrderAtoms(mol)
    oechem.OECanonicalOrderBonds(mol)
    oechem.OEClearAromaticFlags(mol)
    oechem.OEKekulize(mol)

    oechem.OEAssignAromaticFlags(mol, aro_model)


def generate_tautomer_options(
    max_generated: int = 4096,
    rank: bool = True,
    warts: bool = False,
    level: int = 0,
    carbon_hybrid: bool = True,
    save_stereo: bool = False,
    max_zone_size: int = 35,
    max_tautomeric_atoms: int = 70,
    max_time: int = 120,
    kekule: bool = False,
    clear_coords: bool = False,
    hybrid_level: int = oequacpac.OEHybridizationLevel_EnolEnamine,
    max_to_return: int = 256,
    racemic_type: int = oequacpac.OERacemicType_LocalSampled,
) -> oequacpac.OETautomerOptions:
    """
    Generates an oequacpac.OETautomerOptions object from the given settings.

    Args:
        max_generated: int = 4096
            Maximum number of tautomers that may be generated for each input molecule.
        rank: bool = True
            Whether or not tautomers are ranked and ordered before being returned.
        warts: bool = False
            Whether to set warts application of {title}_1, {title}_2, etc., on tautomers.
        level: int = 0
            Level for tautomerization. Higher levels allow less likely atomic transitions to occur.
        carbon_hybrid: bool = True
            Setting for whether or not carbon hybridization changes are allowed. Allowing sp2-sp3 changes
            of carbon atoms can significantly increase computation time.
        save_stereo: bool = False
            Setting for whether atoms and bonds with labeled stereochemistry are fixed or allowed to
            participate in tautomerization.
        max_zone_size: int = 35
            Maximum number of atoms allowed in a continuous tautomerization zone.
        max_tautomeric_atoms: int = 70
            Maximum number of tautomeric atoms allowed in a molecule.
        max_time: 120
            Maximum search time in seconds allowed for a molecule.
        kekule: bool = False
            Setting for whether or not generated tautomers are kekule or have aromaticity perceived.
        clear_coords: bool = False
            Whether input coordinates are cleared.
        hybrid_level: int = oequacpac.OEHybridizationLevel_EnolEnamine
            Setting for the level of hybridization changes allowed when carbon hybridization is allowed.
            Allowing carbon hybridization significantly increase computation time.
        max_to_return: int = 256
            Maximum number of tautomers that may be returned for each input molecule.
        racemic_type: int = oequacpac.OERacemicType_LocalSampled
            Setting for the way to handle loss of stereocenters, if the OETautomerOptions.GetSaveStereo method
            returns false. If the OETautomerOptions.GetSaveStereo method returns true, this option is irrelevant.

    Returns:
        oequacpac.OETautomerOptions
    """
    options = oequacpac.OETautomerOptions(
        max_generated,
        rank,
        warts,
        level,
        carbon_hybrid,
        save_stereo,
        max_zone_size,
        max_tautomeric_atoms,
        max_time,
        kekule,
        clear_coords,
        hybrid_level,
        max_to_return,
        racemic_type,
    )

    return options



from dataclasses import field
from typing import Literal

from numpy.random import RandomState
from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import (
    SKLearnClassificationMixin,
    SKLearnModelBase,
)

try:
    from sklearn.neural_network import MLPClassifier as SKMLPClassifier
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
Multi-layer Perceptron classifier.

This model optimizes the log-loss function using LBFGS or stochastic gradient descent.

Notes
-----
MLPClassifier trains iteratively since at each time step
the partial derivatives of the loss function with respect to the model
parameters are computed to update the parameters.
It can also have a regularization term added to the loss function
that shrinks model parameters to prevent overfitting.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
hidden_layer_sizes : tuple, length = n_layers - 2, default=(100,)
    The ith molflux represents the number of neurons in the ith
    hidden layer.
activation : {'identity', 'logistic', 'tanh', 'relu'}, default='relu'
    Activation function for the hidden layer.
    - 'identity', no-op activation, useful to implement linear bottleneck,
      returns f(x) = x
    - 'logistic', the logistic sigmoid function,
      returns f(x) = 1 / (1 + exp(-x)).
    - 'tanh', the hyperbolic tan function,
      returns f(x) = tanh(x).
    - 'relu', the rectified linear unit function,
      returns f(x) = max(0, x)
solver : {'lbfgs', 'sgd', 'adam'}, default='adam'
    The solver for weight optimization.
    - 'lbfgs' is an optimizer in the family of quasi-Newton methods.
    - 'sgd' refers to stochastic gradient descent.
    - 'adam' refers to a stochastic gradient-based optimizer proposed
      by Kingma, Diederik, and Jimmy Ba
    Note: The default solver 'adam' works pretty well on relatively
    large datasets (with thousands of training samples or more) in terms of
    both training time and validation score.
    For small datasets, however, 'lbfgs' can converge faster and perform
    better.
alpha : float, default=0.0001
    Strength of the L2 regularization term. The L2 regularization term
    is divided by the sample size when added to the loss.
batch_size : int, default='auto'
    Size of minibatches for stochastic optimizers.
    If the solver is 'lbfgs', the classifier will not use minibatch.
    When set to "auto", `batch_size=min(200, n_samples)`.
learning_rate : {'constant', 'invscaling', 'adaptive'}, default='constant'
    Learning rate schedule for weight updates.
    - 'constant' is a constant learning rate given by
      'learning_rate_init'.
    - 'invscaling' gradually decreases the learning rate at each
      time step 't' using an inverse scaling exponent of 'power_t'.
      effective_learning_rate = learning_rate_init / pow(t, power_t)
    - 'adaptive' keeps the learning rate constant to
      'learning_rate_init' as long as training loss keeps decreasing.
      Each time two consecutive epochs fail to decrease training loss by at
      least tol, or fail to increase validation score by at least tol if
      'early_stopping' is on, the current learning rate is divided by 5.
    Only used when ``solver='sgd'``.
learning_rate_init : float, default=0.001
    The initial learning rate used. It controls the step-size
    in updating the weights. Only used when solver='sgd' or 'adam'.
power_t : float, default=0.5
    The exponent for inverse scaling learning rate.
    It is used in updating effective learning rate when the learning_rate
    is set to 'invscaling'. Only used when solver='sgd'.
max_iter : int, default=200
    Maximum number of iterations. The solver iterates until convergence
    (determined by 'tol') or this number of iterations. For stochastic
    solvers ('sgd', 'adam'), note that this determines the number of epochs
    (how many times each data point will be used), not the number of
    gradient steps.
shuffle : bool, default=True
    Whether to shuffle samples in each iteration. Only used when
    solver='sgd' or 'adam'.
random_state : int, RandomState instance, default=None
    Determines random number generation for weights and bias
    initialization, train-test split if early stopping is used, and batch
    sampling when solver='sgd' or 'adam'.
    Pass an int for reproducible results across multiple function calls.
tol : float, default=1e-4
    Tolerance for the optimization. When the loss or score is not improving
    by at least ``tol`` for ``n_iter_no_change`` consecutive iterations,
    unless ``learning_rate`` is set to 'adaptive', convergence is
    considered to be reached and training stops.
verbose : bool, default=False
    Whether to print progress messages to stdout.
warm_start : bool, default=False
    When set to True, reuse the solution of the previous
    call to fit as initialization, otherwise, just erase the
    previous solution.
momentum : float, default=0.9
    Momentum for gradient descent update. Should be between 0 and 1. Only
    used when solver='sgd'.
nesterovs_momentum : bool, default=True
    Whether to use Nesterov's momentum. Only used when solver='sgd' and
    momentum > 0.
early_stopping : bool, default=False
    Whether to use early stopping to terminate training when validation
    score is not improving. If set to true, it will automatically set
    aside 10% of training data as validation and terminate training when
    validation score is not improving by at least tol for
    ``n_iter_no_change`` consecutive epochs. The split is stratified,
    except in a multilabel setting.
    If early stopping is False, then the training stops when the training
    loss does not improve by more than tol for n_iter_no_change consecutive
    passes over the training set.
    Only effective when solver='sgd' or 'adam'.
validation_fraction : float, default=0.1
    The proportion of training data to set aside as validation set for
    early stopping. Must be between 0 and 1.
    Only used if early_stopping is True.
beta_1 : float, default=0.9
    Exponential decay rate for estimates of first moment vector in adam,
    should be in [0, 1). Only used when solver='adam'.
beta_2 : float, default=0.999
    Exponential decay rate for estimates of second moment vector in adam,
    should be in [0, 1). Only used when solver='adam'.
epsilon : float, default=1e-8
    Value for numerical stability in adam. Only used when solver='adam'.
n_iter_no_change : int, default=10
    Maximum number of epochs to not meet ``tol`` improvement.
    Only effective when solver='sgd' or 'adam'.
max_fun : int, default=15000
    Only used when solver='lbfgs'. Maximum number of loss function calls.
    The solver iterates until convergence (determined by 'tol'), number
    of iterations reaches max_iter, or this number of loss function calls.
    Note that number of loss function calls will be greater than or equal
    to the number of iterations for the `MLPClassifier`.
"""

ActivationT = Literal["identity", "logistic", "tanh", "relu"]
SolverT = Literal["lbfgs", "sgd", "adam"]
LearningRateT = Literal["constant", "invscaling", "adaptive"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class MLPClassifierConfig(ModelConfig):
    hidden_layer_sizes: list[int] = field(
        default_factory=lambda: [
            100,
        ],
    )
    activation: ActivationT = "relu"
    solver: SolverT = "adam"
    alpha: float = 0.0001
    batch_size: int | Literal["auto"] = "auto"
    learning_rate: LearningRateT = "constant"
    learning_rate_init: float = 0.001
    power_t: float = 0.5
    max_iter: int = 200
    shuffle: bool = True
    random_state: None | int | RandomState = None
    tol: float = 1e-4
    verbose: bool = False
    warm_start: bool = False
    momentum: float = 0.9
    nesterovs_momentum: bool = True
    early_stopping: bool = False
    validation_fraction: float = 0.1
    beta_1: float = 0.9
    beta_2: float = 0.999
    epsilon: float = 1e-8
    n_iter_no_change: int = 10
    max_fun: int = 15000


class MLPClassifier(SKLearnClassificationMixin, SKLearnModelBase[MLPClassifierConfig]):
    @property
    def _config_builder(self) -> type[MLPClassifierConfig]:
        return MLPClassifierConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SKMLPClassifier:
        config = self.model_config
        return SKMLPClassifier(
            hidden_layer_sizes=tuple(config.hidden_layer_sizes),
            activation=config.activation,
            solver=config.solver,
            alpha=config.alpha,
            batch_size=config.batch_size,
            learning_rate=config.learning_rate,
            learning_rate_init=config.learning_rate_init,
            power_t=config.power_t,
            max_iter=config.max_iter,
            shuffle=config.shuffle,
            random_state=config.random_state,
            tol=config.tol,
            verbose=config.verbose,
            warm_start=config.warm_start,
            momentum=config.momentum,
            nesterovs_momentum=config.nesterovs_momentum,
            early_stopping=config.early_stopping,
            validation_fraction=config.validation_fraction,
            beta_1=config.beta_1,
            beta_2=config.beta_2,
            epsilon=config.epsilon,
            n_iter_no_change=config.n_iter_no_change,
            max_fun=config.max_fun,
        )



from molflux.datasets.catalogue import fill_catalogue, list_datasets
from molflux.datasets.featurisation import featurise_dataset
from molflux.datasets.io import load_dataset_from_store, save_dataset_to_store
from molflux.datasets.load import (
    load_dataset,
    load_dataset_builder,
    load_from_dict,
    load_from_dicts,
    load_from_yaml,
)
from molflux.datasets.splitting import split_dataset

# Register all plugins at package import time (to fill catalogue)
fill_catalogue()



import pathlib
import tempfile

import nox

SUPPORTED_PYTHON_VERSIONS = ["3.10", "3.11"]

nox.options.envdir = ".cache/nox"
nox.options.reuse_existing_virtualenvs = False
nox.options.default_venv_backend = "uv"
nox.options.sessions = [
    "formatting_check-3.11",
    "typing_check-3.11",
    "tests_run_latest-3.11",
]

SOURCE_FILES = ("src/", "tests/", "noxfile.py")

# Artefacts folders
PINNED_VERSIONS = "pinned-versions"
COVERAGE_DIR = ".coverage"

TEST_REPORTS_DIR = "test-reports"

# all external utility tools used by our nox sessions
BUILD_TOOLS = ["build"]
COVERAGE_TOOLS = [
    "coverage[toml]",
    "coverage-badge",
    "setuptools",  # setuptools required with uv as coverage-badge imports pkg_resources
]
RUFF_VERSION = "ruff~=0.4"

SUBMODULE_EXTRAS: dict[str, dict[str, list | None]] = {
    "core": {
        "doctests_target_dir_core": [
            # "src/molflux/core"
        ],
        "tests_target_dir_core": ["tests/core/core"],
        "doctests_target_dir_extras": None,
        "tests_target_dir_extras": None,
        "extras": [None],
    },
    "datasets": {
        "doctests_target_dir_core": [
            # "src/molflux/datasets"
        ],
        "tests_target_dir_core": ["tests/datasets/core", "tests/datasets/plugins/core"],
        "doctests_target_dir_extras": [
            # "src/molflux/datasets/builders"
        ],
        "tests_target_dir_extras": ["tests/datasets/plugins"],
        "extras": [None, "openeye", "rdkit", "tdc"],
    },
    "features": {
        "doctests_target_dir_core": [
            # "src/molflux/features",
            # "src/molflux/features/representations/core",
        ],
        "tests_target_dir_core": ["tests/features/core", "tests/features/plugins/core"],
        "doctests_target_dir_extras": [
            # "src/molflux/features/representations"
        ],
        "tests_target_dir_extras": ["tests/features/plugins"],
        "extras": [None, "openeye", "rdkit"],
    },
    "splits": {
        "doctests_target_dir_core": [
            # "src/molflux/splits",
            # "src/molflux/splits/strategies/core",
        ],
        "tests_target_dir_core": ["tests/splits/core", "tests/splits/plugins/core"],
        "doctests_target_dir_extras": [
            # "src/molflux/splits/strategies"
        ],
        "tests_target_dir_extras": ["tests/splits/plugins"],
        "extras": [None, "openeye", "rdkit"],
    },
    "modelzoo": {
        "doctests_target_dir_core": [
            # "src/molflux/modelzoo",
            # "src/molflux/modelzoo/models/core",
        ],
        "tests_target_dir_core": ["tests/modelzoo/core", "tests/modelzoo/plugins/core"],
        "doctests_target_dir_extras": [
            # "src/molflux/modelzoo/models"
        ],
        "tests_target_dir_extras": ["tests/modelzoo/plugins"],
        "extras": [
            None,
            "catboost",
            "ensemble",
            "lightning",
            "mapie",
            "pyod",
            "pystan",
            "sklearn",
            "xgboost",
        ],
    },
    "metrics": {
        "doctests_target_dir_core": [
            # "src/molflux/metrics"
        ],
        "tests_target_dir_core": ["tests/metrics/core", "tests/metrics/plugins/core"],
        "doctests_target_dir_extras": None,
        "tests_target_dir_extras": None,
        "extras": [None],
    },
}

EXTRAS: list[str | None] = []
for v in SUBMODULE_EXTRAS.values():
    EXTRAS += v["extras"]  # type: ignore

EXTRAS = list(set(EXTRAS))

# skip openeye in github actions (cannot pip install)
EXTRAS.remove("openeye")

INVERTED_SUBMODULE_EXTRAS = {}
for extra in EXTRAS:
    doctests_target_dir_core = []
    tests_target_dir_core = []
    doctests_target_dir_extras = []
    tests_target_dir_extras = []
    for v in SUBMODULE_EXTRAS.values():
        if extra in v["extras"]:  # type: ignore
            if v["doctests_target_dir_core"]:
                doctests_target_dir_core += v["doctests_target_dir_core"]
            if v["tests_target_dir_core"]:
                tests_target_dir_core += v["tests_target_dir_core"]
            if v["doctests_target_dir_extras"]:
                doctests_target_dir_extras += v["doctests_target_dir_extras"]
            if v["tests_target_dir_extras"]:
                tests_target_dir_extras += v["tests_target_dir_extras"]

    INVERTED_SUBMODULE_EXTRAS[extra] = {
        "doctests_target_dir_core": doctests_target_dir_core,
        "tests_target_dir_core": tests_target_dir_core,
        "doctests_target_dir_extras": doctests_target_dir_extras,
        "tests_target_dir_extras": tests_target_dir_extras,
    }


def resolve_lockfile_path(
    python_version: str,
    extra: str | None = None,
    rootdir: str = PINNED_VERSIONS,
) -> pathlib.Path:
    """Resolves the expected lockfile path for a given python version and extra."""
    lockfile_name = f"lockfile.{extra or 'core'}.txt"
    return pathlib.Path(rootdir) / python_version / lockfile_name


def resolve_coverage_datafile_path(
    python_version: str,
    extra: str | None = None,
) -> pathlib.Path:
    """Resolves the expected coverage data_file path for a given python version and extra."""
    coverage_datafile_name = f".coverage.{extra or 'core'}"
    return pathlib.Path(COVERAGE_DIR) / python_version / coverage_datafile_name


def resolve_junitxml_path(
    python_version: str,
    extra: str | None = None,
) -> pathlib.Path:
    """Resolves the output pytest junitxml reports path for a given python version and extra."""
    junitxml_report_name = f".junitxml.{extra or 'core'}.xml"
    return pathlib.Path(TEST_REPORTS_DIR) / python_version / junitxml_report_name


@nox.session(python=SUPPORTED_PYTHON_VERSIONS)
def coverage_report(session: nox.Session) -> None:
    """Generate coverage reports.

    This session is usually triggered following a pytest coverage session generating
    the coverage data files to build the reports on. The directory containing
    the coverage data files should be provided as posarg.

    Examples:
        nox -s coverage_report-3.11 -- .coverage/3.11
    """

    session.install(*COVERAGE_TOOLS)

    # Combine coverage output data_files
    datafiles_dir = session.posargs[0]
    merged_data_file = f"{datafiles_dir}/.coverage"
    session.run(
        "coverage",
        "combine",
        "--data-file",
        merged_data_file,
        "--keep",
        datafiles_dir,
    )

    # Output coverage reports in same folder (for convenience)
    coverage_reports_dir = datafiles_dir
    session.run("coverage", "report", "--data-file", merged_data_file)
    session.run(
        "coverage",
        "html",
        "--data-file",
        merged_data_file,
        "-d",
        f"{coverage_reports_dir}/html",
    )
    session.run(
        "coverage",
        "xml",
        "--data-file",
        merged_data_file,
        "-o",
        f"{coverage_reports_dir}/coverage.xml",
    )
    session.run(
        "coverage",
        "json",
        "--data-file",
        merged_data_file,
        "-o",
        f"{coverage_reports_dir}/coverage.json",
    )

    session.notify(f"coverage_build_badge-{session.python}", [coverage_reports_dir])


@nox.session(python=SUPPORTED_PYTHON_VERSIONS)
def coverage_build_badge(session: nox.Session) -> None:
    """Generate a coverage badge.

    This session is usually triggered following a pytest coverage session generating
    the coverage data files for which to build the badge. The directory containing
    the final .coverage data file should be provided as posarg.

    Examples:
        nox -s coverage_build_badge-3.11 -- .coverage/3.11
    """

    # coverage-badge only works from the same directory where the .coverage
    # data file is located.
    data_file_dir = session.posargs[0]
    session.chdir(data_file_dir)

    badge_filename = "coverage.svg"

    # cleanup old badge
    session.run("rm", "-rf", badge_filename, external=True)

    session.install(*COVERAGE_TOOLS)
    session.run("coverage-badge", "-o", badge_filename)


@nox.session(python=SUPPORTED_PYTHON_VERSIONS)
def dist_build(session: nox.Session) -> None:
    """Build distributions (sdist and wheel).

    The distribution packages are built using PyPA's `build`_ build frontend.
    This is the recommended way of building python packages, avoiding direct
    calls to the build backend. Legacy calls like ``$ python setup.py build``
    are now deprecated.

    .. _build:
            https://pypa-build.readthedocs.io/en/latest/

    Examples:
        nox -s dist_build-3.11
    """

    session.run("rm", "-rf", "dist", external=True)
    session.install(*BUILD_TOOLS)
    session.run("python", "-m", "build")


@nox.session(python=SUPPORTED_PYTHON_VERSIONS)
def docs_build(session: nox.Session) -> None:
    """Build sphinx documentation and API docs.

    Examples:
        nox -s docs_build-3.11
    """

    lockfile_path = resolve_lockfile_path(python_version=session.python)
    session.install(".[docs]", "--constraint", lockfile_path)

    # # Build API docs
    # apidoc_cmd = "sphinx-apidoc -f -o docs/source/pages/reference/api src/molflux --implicit-namespaces"
    # session.run(*apidoc_cmd.split(" "))

    # wipe artefacts from previous runs, in case there are any
    session.run("rm", "-rf", "docs/build/html", external=True)

    # -a -E flags make sure things are built from scratch
    build_cmd = "sphinx-build -a -E docs/source/ docs/build/html"

    # # Run doctests in the documentation
    # session.run(*build_cmd.split(" "), "-b", "doctest")

    # Build HTML pages
    session.run(*build_cmd.split(" "), "-b", "html")


@nox.session(python=SUPPORTED_PYTHON_VERSIONS)
def formatting_check(session: nox.Session) -> None:
    """Check codebase formatting.

    Examples:
        nox -s formatting_check-3.11
    """
    session.install(RUFF_VERSION)
    session.run("ruff", "check")


@nox.session(python=SUPPORTED_PYTHON_VERSIONS)
def formatting_fix(session: nox.Session) -> None:
    """Fix codebase formatting.

    Examples:
        nox -s formatting_fix-3.11
    """
    session.install(RUFF_VERSION)
    session.run("ruff", "format")


@nox.session(python=SUPPORTED_PYTHON_VERSIONS)
def linting_check(session: nox.Session) -> None:
    """Check codebase lint quality.

    Examples:
        nox -s linting_check-3.11
    """
    session.install(RUFF_VERSION)
    session.run("ruff", "check", *session.posargs, *SOURCE_FILES)


@nox.session(python=SUPPORTED_PYTHON_VERSIONS)
def linting_fix(session: nox.Session) -> None:
    """Fix codebase lint quality where possible.

    Examples:
        nox -s linting_fix-3.11
    """
    session.install(RUFF_VERSION)
    session.run("ruff", "check", "--fix", *session.posargs, *SOURCE_FILES)


@nox.session(python=SUPPORTED_PYTHON_VERSIONS)
def typing_check(session: nox.Session) -> None:
    """Check codebase type annotations.

    Examples:
        nox -s typing_check-3.11
        nox -s typing_check-3.11 -- --python-version 3.11
    """

    lockfile_path = resolve_lockfile_path(python_version=session.python)
    session.install(".[tests,typing]", "--constraint", lockfile_path)
    session.run("mypy", *session.posargs)


def generate_lockfile(
    session: nox.Session,
    extra: str | None,
    lockfile_path: pathlib.Path,
) -> None:
    """Generates a package dependencies' lockfile.

    Args:
        session: The nox.Session to use for the operation.
        extra: The name of an additional specific package extra to take into
            account when resolving dependencies. If not None, this will be used
            in addition to the usual 'docs' and 'tests' extras.
        lockfile_path: The path where to output the generated lockfile.
    """

    session.install("--upgrade", "pip", "pip-tools", "uv")

    package_extras = f"docs,tests,{extra}" if extra else "docs,tests"
    lockfile_path.parent.mkdir(parents=True, exist_ok=True)
    session.run(
        "uv",
        "pip",
        "compile",
        *[f"--extra={x}" for x in package_extras.split(",")],
        "--strip-extras",
        "--no-emit-index-url",
        "pyproject.toml",
        "-o",
        str(lockfile_path),
        "--upgrade",
        env={"UV_CUSTOM_COMPILE_COMMAND": f"nox -s {session.name}"},
    )
    print(f"Lockfile generated at {str(lockfile_path)!r} ✨")


@nox.session(python=SUPPORTED_PYTHON_VERSIONS)
@nox.parametrize("extra", EXTRAS)
def dependencies_pin(session: nox.Session, extra: str | None) -> None:
    """Generate pinned dependencies lockfiles.

    Examples:
        nox -s dependencies_pin-3.11
        nox -s "dependencies_pin-3.11(extra=None)"
        nox -s "dependencies_pin-3.11(extra='sklearn')"
    """
    output_lockfile_path = resolve_lockfile_path(
        python_version=session.python,
        extra=extra,
    )
    generate_lockfile(session, extra=extra, lockfile_path=output_lockfile_path)


def run_tests(
    session: nox.Session,
    *args: str,
    extra: str | None,
    lockfile_path: pathlib.Path,
    notify: bool,
) -> None:
    """Runs tests.

    This includes running code snippets in our source code docstrings.

    Args:
        session: The nox.Session to use for the operation.
        extra: The name of the package extra for which to run tests. If None,
            only core tests will be collected.
        lockfile_path: The path to the lockfile to use for constraining the
            testing environment.
        notify: If True, coverage reporting sessions are queued up on success.
    """

    # Setup which files and tests to target
    if extra:
        # install test dependencies and extra dependencies
        package_extras = ",".join(["tests", extra])

        # all code requiring the extra is in its own dir
        doctests_target_dirs = [
            f"{dir}/{extra}"
            for dir in INVERTED_SUBMODULE_EXTRAS[extra]["doctests_target_dir_extras"]
        ]

        # all tests requiring the extra are in their own dir and
        tests_target_dirs = [
            f"{dir}/{extra}"
            for dir in INVERTED_SUBMODULE_EXTRAS[extra]["tests_target_dir_extras"]
        ]

    else:
        # only install test dependencies
        package_extras = "tests"

        # all code requiring extras in its own subdir - so parse everything
        # except those subdirs (see norecursedirs in pyproject.toml)
        doctests_target_dirs = INVERTED_SUBMODULE_EXTRAS[extra][
            "doctests_target_dir_core"
        ]

        # all tests not requiring extras are in their own dir
        tests_target_dirs = INVERTED_SUBMODULE_EXTRAS[extra]["tests_target_dir_core"]

    # Run tests
    install_cmd = f".[{package_extras}] --constraint {lockfile_path}"
    if extra == "pystan":  # handle pystan SyntaxError on uv backend
        install_cmd += " --compile"
    session.install(*install_cmd.split(" "))

    coverage_datafile_path = resolve_coverage_datafile_path(
        python_version=session.python,
        extra=extra,
    )
    junitxml_path = resolve_junitxml_path(python_version=session.python, extra=extra)
    session.run(
        "coverage",
        "run",
        f"--data-file={coverage_datafile_path}",
        "-m",
        "pytest",
        f"--junitxml={junitxml_path}",
        *args,
        *doctests_target_dirs,
        *tests_target_dirs,
    )

    # for parametrised runs, this will get run once at the end of all of them
    if notify:
        datafiles_dir = str(coverage_datafile_path.parent)
        session.notify(f"coverage_report-{session.python}", [datafiles_dir])


@nox.session(python=SUPPORTED_PYTHON_VERSIONS)
@nox.parametrize("extra", EXTRAS)
def tests_run_latest(session: nox.Session, extra: str | None) -> None:
    """Run tests against latest available dependencies.

    Examples:
        nox -s tests_run_latest-3.11
        nox -s "tests_run_latest-3.11(extra=None)"
        nox -s "tests_run_latest-3.11(extra='sklearn')"
    """
    # Generate a scratch lock file with latest resolved dependencies
    #
    # we could have used session.create_tmp but that sets $TMPDIR which creates
    # problems with multiprocessing code: https://github.com/python/cpython/issues/93852
    with tempfile.TemporaryDirectory() as tmp:
        scratch_output_lockfile_path = resolve_lockfile_path(
            python_version=session.python,
            extra=extra,
            rootdir=tmp,
        )
        generate_lockfile(
            session,
            extra=extra,
            lockfile_path=scratch_output_lockfile_path,
        )

        run_tests(
            session,
            *session.posargs,
            extra=extra,
            lockfile_path=scratch_output_lockfile_path,
            notify=False,
        )


@nox.session(python=SUPPORTED_PYTHON_VERSIONS)
@nox.parametrize("extra", EXTRAS)
def tests_run_pinned(session: nox.Session, extra: str | None) -> None:
    """Run tests against pinned dependencies.

    These should already be present. If not, they can be generated / updated
    by running the `dependencies_pin` session.

    Examples:
        nox -s tests_run_pinned-3.11
        nox -s "tests_run_pinned-3.11(extra=None)"
        nox -s "tests_run_pinned-3.11(extra='sklearn')"
    """
    expected_lockfile_path = resolve_lockfile_path(
        python_version=session.python,
        extra=extra,
    )
    run_tests(
        session,
        *session.posargs,
        extra=extra,
        lockfile_path=expected_lockfile_path,
        notify=True,
    )



from collections.abc import Callable
from dataclasses import asdict
from typing import Any, Literal

from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.models.pyod import (
    PyODClassificationMixin,
    PyODModelBase,
    PyODModelConfig,
)

try:
    from pyod.models.knn import KNN
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("pyod", e) from None


_DESCRIPTION = """
kNN class for outlier detection.
For an observation, its distance to its kth nearest neighbour could be
viewed as the outlying score. It could be viewed as a way to measure
the density. See :cite:`ramaswamy2000efficient,angiulli2002fast` for
details.

Three kNN detectors are supported:
largest: use the distance to the kth neighbour as the outlier score
mean: use the average of all k neighbours as the outlier score
median: use the median of the distance to k neighbours as the outlier score
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
contamination : float in (0., 0.5), optional (default=0.1)
    The amount of contamination of the data set,
    i.e. the proportion of outliers in the data set. Used when fitting to
    define the threshold on the decision function.

n_neighbours : int, optional (default = 5)
    Number of neighbours to use by default for k neighbours queries.

method : str, optional (default='largest')
    {'largest', 'mean', 'median'}

    - 'largest': use the distance to the kth neighbour as the outlier score
    - 'mean': use the average of all k neighbours as the outlier score
    - 'median': use the median of the distance to k neighbours as the
      outlier score

radius : float, optional (default = 1.0)
    Range of parameter space to use by default for `radius_neighbours`
    queries.

leaf_size : int, optional (default = 30)
    Leaf size passed to BallTree. This can affect the
    speed of the construction and query, as well as the memory
    required to store the tree.  The optimal value depends on the
    nature of the problem.

metric : string or callable, default 'minkowski'
    metric to use for distance computation. Any metric from scikit-learn
    or scipy.spatial.distance can be used.

    If metric is a callable function, it is called on each
    pair of instances (rows) and the resulting value recorded. The callable
    should take two arrays as input and return one value indicating the
    distance between them. This works for Scipy's metrics, but is less
    efficient than passing the metric name as a string.

    Distance matrices are not supported.

    Valid values for metric are:

    - from scikit-learn: ['cityblock', 'cosine', 'euclidean', 'l1', 'l2',
      'manhattan']

    - from scipy.spatial.distance: ['braycurtis', 'canberra', 'chebyshev',
      'correlation', 'dice', 'hamming', 'jaccard', 'kulsinski',
      'mahalanobis', 'matching', 'minkowski', 'rogerstanimoto',
      'russellrao', 'seuclidean', 'sokalmichener', 'sokalsneath',
      'sqeuclidean', 'yule']

    See the documentation for scipy.spatial.distance for details on these
    metrics.

p : integer, optional (default = 2)
    Parameter for the Minkowski metric from
    sklearn.metrics.pairwise.pairwise_distances. When p = 1, this is
    equivalent to using manhattan_distance (l1), and euclidean_distance
    (l2) for p = 2. For arbitrary p, minkowski_distance (l_p) is used.
    See http://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.pairwise_distances

metric_params : dict, optional (default = None)
    Additional keyword arguments for the metric function.

n_jobs : int, optional (default = 1)
    The number of parallel jobs to run for neighbours search.
    If ``-1``, then the number of jobs is set to the number of CPU cores.
    Affects only kneighbours and kneighbours_graph methods.
"""

Method = Literal["largest", "mean", "median"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class KNNDetectorConfig(PyODModelConfig):
    contamination: float = 0.1
    n_neighbours: int = 5
    method: Method = "largest"
    radius: float = 1.0
    leaf_size: int = 30
    metric: str | Callable = "minkowski"
    p: int = 2
    metric_params: dict | None = None
    n_jobs: int = 1


class KNNDetector(PyODClassificationMixin, PyODModelBase[KNNDetectorConfig]):
    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    @property
    def _config_builder(self) -> type[KNNDetectorConfig]:
        return KNNDetectorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> KNN:
        config = self.model_config
        return KNN(
            contamination=config.contamination,
            n_neighbors=config.n_neighbours,
            method=config.method,
            radius=config.radius,
            algorithm="ball_tree",
            leaf_size=config.leaf_size,
            metric=config.metric,
            p=config.p,
            metric_params=config.metric_params,
            n_jobs=config.n_jobs,
        )



from collections.abc import Iterable
from typing import Any

from molflux.metrics.catalogue import get_metric_cls
from molflux.metrics.metric import Metric, Metrics
from molflux.metrics.parsers import Spec, dict_parser, yaml_parser
from molflux.metrics.suites.catalogue import get_suite_path
from molflux.metrics.typing import PathLike


def load_metric(name: str, **metric_init_kwargs: Any) -> Metric:
    """Loads a `metrics.Metric` from metric name."""

    # Fetch relevant metric class from the catalogue
    metric_cls = get_metric_cls(metric_name=name)

    # Instantiate metric object
    metric = metric_cls(**metric_init_kwargs)

    return metric


def load_metrics(*names: str, tags: list[str] | None = None) -> Metrics:
    """Loads a `metrics.Metrics` collection from metric names.

    This is a utility functions for loading a number of metrics at once, when
    no further customisation is needed.
    """

    tags = tags or [None for _ in names]  # type:ignore[misc]

    metrics = Metrics()
    for name, tag in zip(names, tags, strict=False):
        metric = load_metric(name=name, tag=tag)
        metrics.add_metric(metric=metric)

    return metrics


def _load_from_spec(spec: Spec) -> Metric:
    """Loads a metric from a validated Spec."""

    # Build metric
    metric = load_metric(name=spec.name, **spec.config)

    # update state
    metric.update_state(**spec.presets)

    return metric


def load_from_dict(dictionary: dict[str, Any]) -> Metric:
    """Loads a metric from a config dict."""

    # Validate dictionary
    spec = dict_parser(dictionary=dictionary)

    return _load_from_spec(spec=spec)


def load_from_dicts(dictionaries: Iterable[dict[str, Any]]) -> Metrics:
    """Loads a collection of metrics from an iterable of dicts."""

    metrics = Metrics()
    for dictionary in dictionaries:
        metric = load_from_dict(dictionary=dictionary)

        metrics.add_metric(metric=metric)

    return metrics


def load_from_yaml(path: PathLike) -> Metrics:
    """Loads a collection of metrics from a yaml config file."""

    specs = yaml_parser(path=path)

    metrics = Metrics()
    for spec in specs:
        metric = _load_from_spec(spec=spec)

        metrics.add_metric(metric=metric)

    return metrics


def load_suite(name: str) -> Metrics:
    """Loads a pre-configured suite of metrics."""
    path = get_suite_path(suite_name=name)
    return load_from_yaml(path=path)
